# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin

from flask_babel import gettext
from .subwizards import Subwizards


class CoreWizardPlugin(octoprint.plugin.AssetPlugin,
                       octoprint.plugin.TemplatePlugin,
                       octoprint.plugin.WizardPlugin,
                       octoprint.plugin.SettingsPlugin,
                       octoprint.plugin.BlueprintPlugin,
                       Subwizards):

	#~~ TemplatePlugin API

	def get_template_configs(self):
		required = self._get_subwizard_attrs("_is_", "_wizard_required")
		names = self._get_subwizard_attrs("_get_", "_wizard_name")
		additional = self._get_subwizard_attrs("_get_", "_additional_wizard_template_data")

		firstrunonly = self._get_subwizard_attrs("_is_", "_wizard_firstrunonly")
		firstrun = self._settings.global_get(["server", "firstRun"])

		if not firstrun:
			required = dict((key, value) for key, value in required.items()
			                if not firstrunonly.get(key, lambda: False)())

		result = list()
		for key, method in required.items():
			if not callable(method):
				continue

			if not method():
				continue

			if not key in names:
				continue

			name = names[key]()
			if not name:
				continue

			config = dict(type="wizard",
			              name=name,
			              template="corewizard_{}_wizard.jinja2".format(key),
			              div="wizard_plugin_corewizard_{}".format(key),
			              suffix="_{}".format(key))
			if key in additional:
				additional_result = additional[key]()
				if additional_result:
					config.update(additional_result)
			result.append(config)

		return result

	#~~ AssetPlugin API

	def get_assets(self):
		if self.is_wizard_required():
			return dict(
				js=["js/corewizard.js"],
				css=["css/corewizard.css"]
			)
		else:
			return dict()

	#~~ WizardPlugin API

	def is_wizard_required(self):
		required = self._get_subwizard_attrs("_is_", "_wizard_required")
		firstrunonly = self._get_subwizard_attrs("_is_", "_wizard_firstrunonly")
		firstrun = self._settings.global_get(["server", "firstRun"])

		if not firstrun:
			required = dict((key, value) for key, value in required.items()
			                if not firstrunonly.get(key, lambda: False)())
		any_required = any(map(lambda m: m(), required.values()))

		return any_required

	def get_wizard_details(self):
		result = dict()

		def add_result(key, method):
			result[key] = method()
		self._get_subwizard_attrs("_get_", "_wizard_details", add_result)

		return result

	def get_wizard_version(self):
		return 3

	#~~ helpers

	def _get_subwizard_attrs(self, start, end, callback=None):
		result = dict()

		for item in dir(self):
			if not item.startswith(start) or not item.endswith(end):
				continue

			key = item[len(start):-len(end)]
			if not key:
				continue

			attr = getattr(self, item)
			if callable(callback):
				callback(key, attr)
			result[key] = attr

		return result


__plugin_name__ = "Core Wizard"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Provides wizard dialogs for core components and functionality"
__plugin_disabling_discouraged__ = gettext("Without this plugin OctoPrint will no longer be able to perform "
                                           "setup steps that might be required after an update.")
__plugin_license__ = "AGPLv3"
__plugin_implementation__ = CoreWizardPlugin()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import request, jsonify, make_response, Response
from werkzeug.exceptions import BadRequest
import re

from past.builtins import long, unicode, basestring

from octoprint.settings import settings, valid_boolean_trues
from octoprint.server import printer, printerProfileManager, NO_CONTENT
from octoprint.server.api import api
from octoprint.server.util.flask import no_firstrun_access, get_json_command_from_request

from octoprint.printer import UnknownScript

from octoprint.access.permissions import Permissions

#~~ Printer


@api.route("/printer", methods=["GET"])
@Permissions.STATUS.require(403)
def printerState():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	# process excludes
	excludes = []
	if "exclude" in request.values:
		excludeStr = request.values["exclude"]
		if len(excludeStr.strip()) > 0:
			excludes = list(filter(lambda x: x in ["temperature", "sd", "state"], map(lambda x: x.strip(), excludeStr.split(","))))

	result = {}

	# add temperature information
	if not "temperature" in excludes:
		processor = lambda x: x
		heated_bed = printerProfileManager.get_current_or_default()["heatedBed"]
		heated_chamber = printerProfileManager.get_current_or_default()["heatedChamber"]
		if not heated_bed and not heated_chamber:
			processor = _keep_tools
		elif not heated_bed:
			processor = _delete_bed
		elif not heated_chamber:
			processor = _delete_chamber

		result.update({"temperature": _get_temperature_data(processor)})

	# add sd information
	if not "sd" in excludes and settings().getBoolean(["feature", "sdSupport"]):
		result.update({"sd": {"ready": printer.is_sd_ready()}})

	# add state information
	if not "state" in excludes:
		state = printer.get_current_data()["state"]
		result.update({"state": state})

	return jsonify(result)


#~~ Tool


@api.route("/printer/tool", methods=["POST"])
@no_firstrun_access
@Permissions.CONTROL.require(403)
def printerToolCommand():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	valid_commands = {
		"select": ["tool"],
		"target": ["targets"],
		"offset": ["offsets"],
		"extrude": ["amount"],
		"flowrate": ["factor"]
	}
	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	validation_regex = re.compile(r"tool\d+")

	tags = {"source:api", "api:printer.tool"}

	##~~ tool selection
	if command == "select":
		tool = data["tool"]
		if not isinstance(tool, basestring) or re.match(validation_regex, tool) is None:
			return make_response("Invalid tool: %s" % tool, 400)
		if not tool.startswith("tool"):
			return make_response("Invalid tool for selection: %s" % tool, 400)

		printer.change_tool(tool, tags=tags)

	##~~ temperature
	elif command == "target":
		targets = data["targets"]

		# make sure the targets are valid and the values are numbers
		validated_values = {}
		for tool, value in targets.items():
			if re.match(validation_regex, tool) is None:
				return make_response("Invalid target for setting temperature: %s" % tool, 400)
			if not isinstance(value, (int, long, float)):
				return make_response("Not a number for %s: %r" % (tool, value), 400)
			validated_values[tool] = value

		# perform the actual temperature commands
		for tool in validated_values.keys():
			printer.set_temperature(tool, validated_values[tool], tags=tags)

	##~~ temperature offset
	elif command == "offset":
		offsets = data["offsets"]

		# make sure the targets are valid, the values are numbers and in the range [-50, 50]
		validated_values = {}
		for tool, value in offsets.items():
			if re.match(validation_regex, tool) is None:
				return make_response("Invalid target for setting temperature: %s" % tool, 400)
			if not isinstance(value, (int, long, float)):
				return make_response("Not a number for %s: %r" % (tool, value), 400)
			if not -50 <= value <= 50:
				return make_response("Offset %s not in range [-50, 50]: %f" % (tool, value), 400)
			validated_values[tool] = value

		# set the offsets
		printer.set_temperature_offset(validated_values)

	##~~ extrusion
	elif command == "extrude":
		if printer.is_printing():
			# do not extrude when a print job is running
			return make_response("Printer is currently printing", 409)

		amount = data["amount"]
		speed = data.get("speed", None)
		if not isinstance(amount, (int, long, float)):
			return make_response("Not a number for extrusion amount: %r" % amount, 400)
		printer.extrude(amount, speed=speed, tags=tags)

	elif command == "flowrate":
		factor = data["factor"]
		if not isinstance(factor, (int, long, float)):
			return make_response("Not a number for flow rate: %r" % factor, 400)
		try:
			printer.flow_rate(factor, tags=tags)
		except ValueError as e:
			return make_response("Invalid value for flow rate: %s" % str(e), 400)

	return NO_CONTENT


@api.route("/printer/tool", methods=["GET"])
@no_firstrun_access
@Permissions.STATUS.require(403)
def printerToolState():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	return jsonify(_get_temperature_data(_keep_tools))


##~~ Heated bed


@api.route("/printer/bed", methods=["POST"])
@no_firstrun_access
@Permissions.CONTROL.require(403)
def printerBedCommand():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	if not printerProfileManager.get_current_or_default()["heatedBed"]:
		return make_response("Printer does not have a heated bed", 409)

	valid_commands = {
		"target": ["target"],
		"offset": ["offset"]
	}
	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	tags = {"source:api", "api:printer.bed"}

	##~~ temperature
	if command == "target":
		target = data["target"]

		# make sure the target is a number
		if not isinstance(target, (int, long, float)):
			return make_response("Not a number: %r" % target, 400)

		# perform the actual temperature command
		printer.set_temperature("bed", target, tags=tags)

	##~~ temperature offset
	elif command == "offset":
		offset = data["offset"]

		# make sure the offset is valid
		if not isinstance(offset, (int, long, float)):
			return make_response("Not a number: %r" % offset, 400)
		if not -50 <= offset <= 50:
			return make_response("Offset not in range [-50, 50]: %f" % offset, 400)

		# set the offsets
		printer.set_temperature_offset({"bed": offset})

	return NO_CONTENT


@api.route("/printer/bed", methods=["GET"])
@no_firstrun_access
@Permissions.STATUS.require(403)
def printerBedState():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	if not printerProfileManager.get_current_or_default()["heatedBed"]:
		return make_response("Printer does not have a heated bed", 409)

	data = _get_temperature_data(_keep_bed)
	if isinstance(data, Response):
		return data
	else:
		return jsonify(data)


##~~ Heated chamber


@api.route("/printer/chamber", methods=["POST"])
@no_firstrun_access
@Permissions.CONTROL.require(403)
def printerChamberCommand():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	if not printerProfileManager.get_current_or_default()["heatedChamber"]:
		return make_response("Printer does not have a heated chamber", 409)

	valid_commands = {
		"target": ["target"],
		"offset": ["offset"]
	}
	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	tags = {"source:api", "api:printer.chamber"}

	##~~ temperature
	if command == "target":
		target = data["target"]

		# make sure the target is a number
		if not isinstance(target, (int, long, float)):
			return make_response("Not a number: %r" % target, 400)

		# perform the actual temperature command
		printer.set_temperature("chamber", target, tags=tags)

	##~~ temperature offset
	elif command == "offset":
		offset = data["offset"]

		# make sure the offset is valid
		if not isinstance(offset, (int, long, float)):
			return make_response("Not a number: %r" % offset, 400)
		if not -50 <= offset <= 50:
			return make_response("Offset not in range [-50, 50]: %f" % offset, 400)

		# set the offsets
		printer.set_temperature_offset({"chamber": offset})

	return NO_CONTENT


@api.route("/printer/chamber", methods=["GET"])
@no_firstrun_access
@Permissions.STATUS.require(403)
def printerChamberState():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	if not printerProfileManager.get_current_or_default()["heatedChamber"]:
		return make_response("Printer does not have a heated chamber", 409)

	data = _get_temperature_data(_keep_chamber)
	if isinstance(data, Response):
		return data
	else:
		return jsonify(data)


##~~ Print head


@api.route("/printer/printhead", methods=["POST"])
@no_firstrun_access
@Permissions.CONTROL.require(403)
def printerPrintheadCommand():
	valid_commands = {
		"jog": [],
		"home": ["axes"],
		"feedrate": ["factor"]
	}
	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	if not printer.is_operational() or (printer.is_printing() and command != "feedrate"):
		# do not jog when a print job is running or we don't have a connection
		return make_response("Printer is not operational or currently printing", 409)

	tags = {"source:api", "api:printer.printhead"}

	valid_axes = ["x", "y", "z"]
	##~~ jog command
	if command == "jog":
		# validate all jog instructions, make sure that the values are numbers
		validated_values = {}
		for axis in valid_axes:
			if axis in data:
				value = data[axis]
				if not isinstance(value, (int, long, float)):
					return make_response("Not a number for axis %s: %r" % (axis, value), 400)
				validated_values[axis] = value

		absolute = "absolute" in data and data["absolute"] in valid_boolean_trues
		speed = data.get("speed", None)

		# execute the jog commands
		printer.jog(validated_values, relative=not absolute, speed=speed, tags=tags)

	##~~ home command
	elif command == "home":
		validated_values = []
		axes = data["axes"]
		for axis in axes:
			if not axis in valid_axes:
				return make_response("Invalid axis: %s" % axis, 400)
			validated_values.append(axis)

		# execute the home command
		printer.home(validated_values, tags=tags)

	elif command == "feedrate":
		factor = data["factor"]
		if not isinstance(factor, (int, long, float)):
			return make_response("Not a number for feed rate: %r" % factor, 400)
		try:
			printer.feed_rate(factor, tags=tags)
		except ValueError as e:
			return make_response("Invalid value for feed rate: %s" % str(e), 400)

	return NO_CONTENT


##~~ SD Card


@api.route("/printer/sd", methods=["POST"])
@no_firstrun_access
@Permissions.CONTROL.require(403)
def printerSdCommand():
	if not settings().getBoolean(["feature", "sdSupport"]):
		return make_response("SD support is disabled", 404)

	if not printer.is_operational() or printer.is_printing() or printer.is_paused():
		return make_response("Printer is not operational or currently busy", 409)

	valid_commands = {
		"init": [],
		"refresh": [],
		"release": []
	}
	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	tags = {"source:api", "api:printer.sd"}

	if command == "init":
		printer.init_sd_card(tags=tags)
	elif command == "refresh":
		printer.refresh_sd_files(tags=tags)
	elif command == "release":
		printer.release_sd_card(tags=tags)

	return NO_CONTENT


@api.route("/printer/sd", methods=["GET"])
@no_firstrun_access
@Permissions.STATUS.require(403)
def printerSdState():
	if not settings().getBoolean(["feature", "sdSupport"]):
		return make_response("SD support is disabled", 404)

	return jsonify(ready=printer.is_sd_ready())


##~~ Commands


@api.route("/printer/command", methods=["POST"])
@no_firstrun_access
@Permissions.CONTROL.require(403)
def printerCommand():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content type JSON", 400)

	try:
		data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if data is None:
		return make_response("Malformed JSON body in request", 400)

	if "command" in data and "commands" in data:
		return make_response("'command' and 'commands' are mutually exclusive", 400)
	elif ("command" in data or "commands" in data) and "script" in data:
		return make_response("'command'/'commands' and 'script' are mutually exclusive", 400)
	elif not ("command" in data or "commands" in data or "script" in data):
		return make_response("Need one of 'command', 'commands' or 'script'", 400)

	parameters = dict()
	if "parameters" in data:
		parameters = data["parameters"]

	tags = {"source:api", "api:printer.command"}

	if "command" in data or "commands" in data:
		if "command" in data:
			commands = [data["command"]]
		else:
			if not isinstance(data["commands"], (list, tuple)):
				return make_response("'commands' needs to be a list", 400)
			commands = data["commands"]

		commandsToSend = []
		for command in commands:
			commandToSend = command
			if len(parameters) > 0:
				commandToSend = command % parameters
			commandsToSend.append(commandToSend)

		printer.commands(commandsToSend, tags=tags)

	elif "script" in data:
		script_name = data["script"]
		context = dict(parameters=parameters)
		if "context" in data:
			context["context"] = data["context"]

		try:
			printer.script(script_name, context=context, tags=tags)
		except UnknownScript:
			return make_response("Unknown script: {script_name}".format(**locals()), 404)

	return NO_CONTENT

@api.route("/printer/command/custom", methods=["GET"])
@no_firstrun_access
@Permissions.CONTROL.require(403)
def getCustomControls():
	customControls = settings().get(["controls"])
	return jsonify(controls=customControls)


def _get_temperature_data(preprocessor):
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	tempData = printer.get_current_temperatures()

	if "history" in request.values and request.values["history"] in valid_boolean_trues:
		tempHistory = printer.get_temperature_history()

		limit = 300
		if "limit" in request.values and unicode(request.values["limit"]).isnumeric():
			limit = int(request.values["limit"])

		history = list(tempHistory)
		limit = min(limit, len(history))

		tempData.update({
			"history": list(map(lambda x: preprocessor(x), history[-limit:]))
		})

	return preprocessor(tempData)


def _keep_tools(x):
	return _delete_from_data(x, lambda k: not k.startswith("tool") and k != "history")


def _keep_bed(x):
	return _delete_from_data(x, lambda k: k != "bed" and k != "history")

def _delete_bed(x):
	return _delete_from_data(x, lambda k: k == "bed")

def _keep_chamber(x):
	return _delete_from_data(x, lambda k: k != "chamber" and k != "history")

def _delete_chamber(x):
	return _delete_from_data(x, lambda k: k == "chamber")


def _delete_from_data(x, key_matcher):
	data = dict(x)
	# must make list of keys first to avoid
	# RuntimeError: dictionary changed size during iteration
	for k in list(data.keys()):
		if key_matcher(k):
			del data[k]
	return data



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

def get_latest(target, check, online=True):
	current_version = check.get("current_version", "1.0.0")

	information = dict(local=dict(name=current_version, value=current_version),
	                   remote=dict(name=current_version, value=current_version))

	return information, True



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import os
import mimetypes
import re
import sys

import tornado
import tornado.web
import tornado.gen
import tornado.escape
import tornado.httputil
import tornado.httpserver
import tornado.httpclient
import tornado.http1connection
import tornado.iostream
import tornado.tcpserver
import tornado.util

import octoprint.util

from past.builtins import unicode

from . import PY3

def fix_json_encode():
	"""
	This makes tornado.escape.json_encode use octoprint.util.JsonEncoding.encode as fallback in order to allow
	serialization of globally registered types like frozendict and others.
	"""

	from octoprint.util.json import JsonEncoding
	import json

	def fixed_json_encode(value):
		return json.dumps(value, default=JsonEncoding.encode, allow_nan=False).replace("</", "<\\/")

	import tornado.escape
	tornado.escape.json_encode = fixed_json_encode


#~~ More sensible logging


class RequestlessExceptionLoggingMixin(tornado.web.RequestHandler):

	LOG_REQUEST = False

	def log_exception(self, typ, value, tb, *args, **kwargs):
		if isinstance(value, tornado.web.HTTPError):
			if value.log_message:
				format = "%d %s: " + value.log_message
				args = ([value.status_code, self._request_summary()] +
				        list(value.args))
				tornado.web.gen_log.warning(format, *args)
		else:
			if self.LOG_REQUEST:
				tornado.web.app_log.error("Uncaught exception %s\n%r", self._request_summary(),
				                          self.request, exc_info=(typ, value, tb))
			else:
				tornado.web.app_log.error("Uncaught exception %s", self._request_summary(),
				                          exc_info=(typ, value, tb))


#~~ CORS support


class CorsSupportMixin(tornado.web.RequestHandler):
	"""
	`tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ mixin that
	makes sure to set CORS headers similarly to the Flask backed API endpoints.
	"""

	ENABLE_CORS = False

	def set_default_headers(self):
		origin = self.request.headers.get("Origin")
		if self.request.method != "OPTIONS" and origin and self.ENABLE_CORS:
			self.set_header("Access-Control-Allow-Origin", origin)

	@tornado.web.asynchronous
	def options(self, *args, **kwargs):
		if self.ENABLE_CORS:
			origin = self.request.headers.get("Origin")
			method = self.request.headers.get("Access-Control-Request-Method")

			# Allow the origin which made the XHR
			self.set_header("Access-Control-Allow-Origin", origin)
			# Allow the actual method
			self.set_header("Access-Control-Allow-Methods", method)
			# Allow for 10 seconds
			self.set_header("Access-Control-Max-Age", "10")

			# 'preflight' request contains the non-standard headers the real request will have (like X-Api-Key)
			custom_headers = self.request.headers.get("Access-Control-Request-Headers")
			if custom_headers is not None:
				self.set_header("Access-Control-Allow-Headers", custom_headers)

		self.set_status(204)
		self.finish()


#~~ WSGI middleware


@tornado.web.stream_request_body
class UploadStorageFallbackHandler(RequestlessExceptionLoggingMixin,
                                   CorsSupportMixin):
	"""
	A ``RequestHandler`` similar to ``tornado.web.FallbackHandler`` which fetches any files contained in the request bodies
	of content type ``multipart``, stores them in temporary files and supplies the ``fallback`` with the file's ``name``,
	``content_type``, ``path`` and ``size`` instead via a rewritten body.

	Basically similar to what the nginx upload module does.

	Basic request body example:

	.. code-block:: none

	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="file"; filename="test.gcode"
	    Content-Type: application/octet-stream

	    ...
	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="apikey"

	    my_funny_apikey
	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="select"

	    true
	    ------WebKitFormBoundarypYiSUx63abAmhT5C--

	That would get turned into:

	.. code-block:: none

	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="apikey"

	    my_funny_apikey
	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="select"

	    true
	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="file.path"
	    Content-Type: text/plain; charset=utf-8

	    /tmp/tmpzupkro
	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="file.name"
	    Content-Type: text/plain; charset=utf-8

	    test.gcode
	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="file.content_type"
	    Content-Type: text/plain; charset=utf-8

	    application/octet-stream
	    ------WebKitFormBoundarypYiSUx63abAmhT5C
	    Content-Disposition: form-data; name="file.size"
	    Content-Type: text/plain; charset=utf-8

	    349182
	    ------WebKitFormBoundarypYiSUx63abAmhT5C--

	The underlying application can then access the contained files via their respective paths and just move them
	where necessary.
	"""

	BODY_METHODS = ("POST", "PATCH", "PUT")
	""" The request methods that may contain a request body. """

	def initialize(self, fallback, file_prefix="tmp", file_suffix="", path=None, suffixes=None):
		if not suffixes:
			suffixes = dict()

		self._fallback = fallback
		self._file_prefix = file_prefix
		self._file_suffix = file_suffix
		self._path = path

		self._suffixes = dict((key, key) for key in ("name", "path", "content_type", "size"))
		for suffix_type, suffix in suffixes.items():
			if suffix_type in self._suffixes and suffix is not None:
				self._suffixes[suffix_type] = suffix

		# multipart boundary
		self._multipart_boundary = None

		# Parts, files and values will be stored here
		self._parts = dict()
		self._files = []

		# Part currently being processed
		self._current_part = None

		# content type of request body
		self._content_type = None

		# bytes left to read according to content_length of request body
		self._bytes_left = 0

		# buffer needed for identifying form data parts
		self._buffer = b""

		# buffer for new body
		self._new_body = b""

		# logger
		self._logger = logging.getLogger(__name__)

	def prepare(self):
		"""
		Prepares the processing of the request. If it's a request that may contain a request body (as defined in
		:attr:`UploadStorageFallbackHandler.BODY_METHODS`) prepares the multipart parsing if content type fits. If it's a
		body-less request, just calls the ``fallback`` with an empty body and finishes the request.
		"""
		if self.request.method in UploadStorageFallbackHandler.BODY_METHODS:
			self._bytes_left = self.request.headers.get("Content-Length", 0)
			self._content_type = self.request.headers.get("Content-Type", None)

			# request might contain a body
			if self.is_multipart():
				if not self._bytes_left:
					# we don't support requests without a content-length
					raise tornado.web.HTTPError(411, log_message="No Content-Length supplied")

				# extract the multipart boundary
				fields = self._content_type.split(";")
				for field in fields:
					k, sep, v = field.strip().partition("=")
					if k == "boundary" and v:
						if v.startswith('"') and v.endswith('"'):
							self._multipart_boundary = tornado.escape.utf8(v[1:-1])
						else:
							self._multipart_boundary = tornado.escape.utf8(v)
						break
				else:
					# RFC2046 section 5.1 (as referred to from RFC 7578) defines the boundary
					# parameter as mandatory for multipart requests:
					#
					#     The only mandatory global parameter for the "multipart" media type is
					#     the boundary parameter, which consists of 1 to 70 characters [...]
					#
					# So no boundary? 400 Bad Request
					raise tornado.web.HTTPError(400, log_message="No multipart boundary supplied")
		else:
			self._fallback(self.request, b"")
			self._finished = True

	def data_received(self, chunk):
		"""
		Called by Tornado on receiving a chunk of the request body. If request is a multipart request, takes care of
		processing the multipart data structure via :func:`_process_multipart_data`. If not, just adds the chunk to
		internal in-memory buffer.

		:param chunk: chunk of data received from Tornado
		"""

		data = self._buffer + chunk
		if self.is_multipart():
			self._process_multipart_data(data)
		else:
			self._buffer = data

	def is_multipart(self):
		"""Checks whether this request is a ``multipart`` request"""
		return self._content_type is not None and self._content_type.startswith("multipart")

	def _process_multipart_data(self, data):
		"""
		Processes the given data, parsing it for multipart definitions and calling the appropriate methods.

		:param data: the data to process as a string
		"""

		# check for boundary
		delimiter = b"--%s" % self._multipart_boundary
		delimiter_loc = data.find(delimiter)
		delimiter_len = len(delimiter)
		end_of_header = -1
		if delimiter_loc != -1:
			# found the delimiter in the currently available data
			delimiter_data_end = 0 if delimiter_loc == 0 else delimiter_loc - 2
			data, self._buffer = data[0:delimiter_data_end], data[delimiter_loc:]
			end_of_header = self._buffer.find(b"\r\n\r\n")
		else:
			# make sure any boundary (with single or double ==) contained at the end of chunk does not get
			# truncated by this processing round => save it to the buffer for next round
			endlen = len(self._multipart_boundary) + 4
			data, self._buffer = data[0:-endlen], data[-endlen:]

		# stream data to part handler
		if data and self._current_part:
				self._on_part_data(self._current_part, data)

		if end_of_header >= 0:
			self._on_part_header(self._buffer[delimiter_len+2:end_of_header])
			self._buffer = self._buffer[end_of_header + 4:]

		if delimiter_loc != -1 and self._buffer.strip() == delimiter + b"--":
			# we saw the last boundary and are at the end of our request
			if self._current_part:
				self._on_part_finish(self._current_part)
				self._current_part = None
			self._buffer = b""
			self._on_request_body_finish()

	def _on_part_header(self, header):
		"""
		Called for a new multipart header, takes care of parsing the header and calling :func:`_on_part` with the
		relevant data, setting the current part in the process.

		:param header: header to parse
		"""

		# close any open parts
		if self._current_part:
			self._on_part_finish(self._current_part)
			self._current_part = None

		header_check = header.find(self._multipart_boundary)
		if header_check != -1:
			self._logger.warning("Header still contained multipart boundary, stripping it...")
			header = header[header_check:]

		# convert to dict
		try:
			header = tornado.httputil.HTTPHeaders.parse(header.decode("utf-8"))
		except UnicodeDecodeError:
			try:
				header = tornado.httputil.HTTPHeaders.parse(header.decode("iso-8859-1"))
			except Exception:
				# looks like we couldn't decode something here neither as UTF-8 nor ISO-8859-1
				self._logger.warning("Could not decode multipart headers in request, should be either UTF-8 or ISO-8859-1")
				self.send_error(400)
				return

		disp_header = header.get("Content-Disposition", "")
		disposition, disp_params = _parse_header(disp_header, strip_quotes=False)

		if disposition != "form-data":
			self._logger.warning("Got a multipart header without form-data content disposition, ignoring that one")
			return
		if not disp_params.get("name"):
			self._logger.warning("Got a multipart header without name, ignoring that one")
			return

		filename = disp_params.get("filename*", None) # RFC 5987 header present?
		if filename is not None:
			try:
				filename = _extended_header_value(filename)
			except Exception:
				# parse error, this is not RFC 5987 compliant after all
				self._logger.warning("extended filename* value {!r} is not RFC 5987 compliant".format(filename))
				self.send_error(400)
				return
		else:
			# no filename* header, just strip quotes from filename header then and be done
			filename = _strip_value_quotes(disp_params.get("filename", None))

		self._current_part = self._on_part_start(_strip_value_quotes(disp_params["name"]),
		                                         header.get("Content-Type", None),
		                                         filename=filename)

	def _on_part_start(self, name, content_type, filename=None):
		"""
		Called for new parts in the multipart stream. If ``filename`` is given creates new ``file`` part (which leads
		to storage of the data as temporary file on disk), if not creates a new ``data`` part (which stores
		incoming data in memory).

		Structure of ``file`` parts:

		* ``name``: name of the part
		* ``filename``: filename associated with the part
		* ``path``: path to the temporary file storing the file's data
		* ``content_type``: content type of the part
		* ``file``: file handle for the temporary file (mode "wb", not deleted on close, will be deleted however after
		  handling of the request has finished in :func:`_handle_method`)

		Structure of ``data`` parts:

		* ``name``: name of the part
		* ``content_type``: content type of the part
		* ``data``: bytes of the part (initialized to an empty string)

		:param name: name of the part
		:param content_type: content type of the part
		:param filename: filename associated with the part.
		:return: dict describing the new part
		"""
		if filename is not None:
			# this is a file
			import tempfile
			handle = tempfile.NamedTemporaryFile(mode="wb", prefix=self._file_prefix, suffix=self._file_suffix, dir=self._path, delete=False)
			return dict(name=tornado.escape.utf8(name),
						filename=tornado.escape.utf8(filename),
						path=tornado.escape.utf8(handle.name),
						content_type=tornado.escape.utf8(content_type),
						file=handle)

		else:
			return dict(name=tornado.escape.utf8(name), content_type=tornado.escape.utf8(content_type), data=b"")

	def _on_part_data(self, part, data):
		"""
		Called when new bytes are received for the given ``part``, takes care of writing them to their storage.

		:param part: part for which data was received
		:param data: data chunk which was received
		"""
		if "file" in part:
			part["file"].write(data)
		else:
			part["data"] += data

	def _on_part_finish(self, part):
		"""
		Called when a part gets closed, takes care of storing the finished part in the internal parts storage and for
		``file`` parts closing the temporary file and storing the part in the internal files storage.

		:param part: part which was closed
		"""
		name = part["name"]
		self._parts[name] = part
		if "file" in part:
			self._files.append(part["path"])
			part["file"].close()
			del part["file"]

	def _on_request_body_finish(self):
		"""
		Called when the request body has been read completely. Takes care of creating the replacement body out of the
		logged parts, turning ``file`` parts into new ``data`` parts.
		"""

		self._new_body = b""
		for name, part in self._parts.items():
			if "filename" in part:
				# add form fields for filename, path, size and content_type for all files contained in the request
				if not "path" in part:
					continue

				parameters = dict(
					name=part["filename"],
					path=part["path"],
					size=str(os.stat(part["path"]).st_size)
				)
				if "content_type" in part:
					parameters["content_type"] = part["content_type"]

				fields = dict((self._suffixes[key], value) for (key, value) in parameters.items())
				for n, p in fields.items():
					if n is None or p is None:
						continue
					key = name + b"." + octoprint.util.to_bytes(n)
					self._new_body += b"--%s\r\n" % self._multipart_boundary
					self._new_body += b"Content-Disposition: form-data; name=\"%s\"\r\n" % key
					self._new_body += b"Content-Type: text/plain; charset=utf-8\r\n"
					self._new_body += b"\r\n"
					self._new_body += octoprint.util.to_bytes(p) + b'\r\n'
			elif "data" in part:
				self._new_body += b"--%s\r\n" % self._multipart_boundary
				value = part["data"]
				self._new_body += b"Content-Disposition: form-data; name=\"%s\"\r\n" % name
				if "content_type" in part and part["content_type"] is not None:
					self._new_body += b"Content-Type: %s\r\n" % part["content_type"]
				self._new_body += b"\r\n"
				self._new_body += value + b"\r\n"
		self._new_body += b"--%s--\r\n" % self._multipart_boundary

	def _handle_method(self, *args, **kwargs):
		"""
		Takes care of defining the new request body if necessary and forwarding
		the current request and changed body to the ``fallback``.
		"""

		# determine which body to supply
		body = b""
		if self.is_multipart():
			# make sure we really processed all data in the buffer
			while len(self._buffer):
				self._process_multipart_data(self._buffer)

			# use rewritten body
			body = self._new_body

		elif self.request.method in UploadStorageFallbackHandler.BODY_METHODS:
			# directly use data from buffer
			body = self._buffer

		# rewrite content length
		self.request.headers["Content-Length"] = len(body)

		try:
			# call the configured fallback with request and body to use
			self._fallback(self.request, body)
			self._headers_written = True
		finally:
			# make sure the temporary files are removed again
			for f in self._files:
				octoprint.util.silent_remove(f)

	# make all http methods trigger _handle_method
	get = _handle_method
	post = _handle_method
	put = _handle_method
	patch = _handle_method
	delete = _handle_method
	head = _handle_method
	options = _handle_method


def _parse_header(line, strip_quotes=True):
	parts = tornado.httputil._parseparam(';' + line)
	key = next(parts)
	pdict = {}
	for p in parts:
		i = p.find('=')
		if i >= 0:
			name = p[:i].strip().lower()
			value = p[i + 1:].strip()
			if strip_quotes:
				value = _strip_value_quotes(value)
			pdict[name] = value
	return key, pdict


def _strip_value_quotes(value):
	if not value:
		return value

	if len(value) >= 2 and value[0] == value[-1] == '"':
		value = value[1:-1]
		value = value.replace('\\\\', '\\').replace('\\"', '"')

	return value


def _extended_header_value(value):
	if not value:
		return value

	if value.lower().startswith("iso-8859-1'") or value.lower().startswith("utf-8'"):
		# RFC 5987 section 3.2
		try:
			from urllib import unquote
		except ImportError:
			from urllib.parse import unquote
		encoding, _, value = value.split("'", 2)
		if PY3:
			return unquote(value, encoding=encoding)
		else:
			return unquote(octoprint.util.to_bytes(value, encoding="iso-8859-1")).decode(encoding)
	else:
		# no encoding provided, strip potentially present quotes and call it a day
		return octoprint.util.to_unicode(_strip_value_quotes(value), encoding="utf-8")


class WsgiInputContainer(object):
	"""
	A WSGI container for use with Tornado that allows supplying the request body to be used for ``wsgi.input`` in the
	generated WSGI environment upon call.

	A ``RequestHandler`` can thus provide the WSGI application with a stream for the request body, or a modified body.

	Example usage:

	.. code-block:: python

	   wsgi_app = octoprint.server.util.WsgiInputContainer(octoprint_app)
	   application = tornado.web.Application([
	       (r".*", UploadStorageFallbackHandler, dict(fallback=wsgi_app),
	   ])

	The implementation logic is basically the same as ``tornado.wsgi.WSGIContainer`` but the ``__call__`` and ``environ``
	methods have been adjusted to allow for an optionally supplied ``body`` argument which is then used for ``wsgi.input``.
	"""

	def __init__(self, wsgi_application, headers=None, forced_headers=None, removed_headers=None):
		self.wsgi_application = wsgi_application

		if headers is None:
			headers = dict()
		if forced_headers is None:
			forced_headers = dict()
		if removed_headers is None:
			removed_headers = []

		self.headers = headers
		self.forced_headers = forced_headers
		self.removed_headers = removed_headers

	def __call__(self, request, body=None):
		"""
		Wraps the call against the WSGI app, deriving the WSGI environment from the supplied Tornado ``HTTPServerRequest``.

		:param request: the ``tornado.httpserver.HTTPServerRequest`` to derive the WSGI environment from
		:param body: an optional body  to use as ``wsgi.input`` instead of ``request.body``, can be a string or a stream
		"""

		data = {}
		response = []

		def start_response(status, response_headers, exc_info=None):
			data["status"] = status
			data["headers"] = response_headers
			return response.append
		app_response = self.wsgi_application(
			WsgiInputContainer.environ(request, body), start_response)
		try:
			response.extend(app_response)
			body = b"".join(response)
		finally:
			if hasattr(app_response, "close"):
				app_response.close()
		if not data:
			raise Exception("WSGI app did not call start_response")

		status_code, reason = data["status"].split(" ", 1)
		status_code = int(status_code)
		headers = data["headers"]
		header_set = set(k.lower() for (k, v) in headers)
		body = tornado.escape.utf8(body)
		if status_code != 304:
			if "content-length" not in header_set:
				headers.append(("Content-Length", str(len(body))))
			if "content-type" not in header_set:
				headers.append(("Content-Type", "text/html; charset=UTF-8"))

		header_set = set(k.lower() for (k, v) in headers)
		for header, value in self.headers.items():
			if header.lower() not in header_set:
				headers.append((header, value))
		for header, value in self.forced_headers.items():
			headers.append((header, value))
		headers = [(header, value) for header, value in headers if not header.lower() in self.removed_headers]

		start_line = tornado.httputil.ResponseStartLine("HTTP/1.1", status_code, reason)
		header_obj = tornado.httputil.HTTPHeaders()
		for key, value in headers:
			header_obj.add(key, value)
		request.connection.write_headers(start_line, header_obj, chunk=body)
		request.connection.finish()
		self._log(status_code, request)

	@staticmethod
	def environ(request, body=None):
		"""
		Converts a ``tornado.httputil.HTTPServerRequest`` to a WSGI environment.

		An optional ``body`` to be used for populating ``wsgi.input`` can be supplied (either a string or a stream). If not
		supplied, ``request.body`` will be wrapped into a ``io.BytesIO`` stream and used instead.

		:param request: the ``tornado.httpserver.HTTPServerRequest`` to derive the WSGI environment from
		:param body: an optional body  to use as ``wsgi.input`` instead of ``request.body``, can be a string or a stream
		"""
		from tornado.wsgi import to_wsgi_str
		import io

		# determine the request_body to supply as wsgi.input
		if body is not None:
			if isinstance(body, (bytes, str, unicode)):
				request_body = io.BytesIO(tornado.escape.utf8(body))
			else:
				request_body = body
		else:
			request_body = io.BytesIO(tornado.escape.utf8(request.body))

		hostport = request.host.split(":")
		if len(hostport) == 2:
			host = hostport[0]
			port = int(hostport[1])
		else:
			host = request.host
			port = 443 if request.protocol == "https" else 80
		environ = {
			"REQUEST_METHOD": request.method,
			"SCRIPT_NAME": "",
			"PATH_INFO": to_wsgi_str(tornado.escape.url_unescape(
				request.path, encoding=None, plus=False)),
			"QUERY_STRING": request.query,
			"REMOTE_ADDR": request.remote_ip,
			"SERVER_NAME": host,
			"SERVER_PORT": str(port),
			"SERVER_PROTOCOL": request.version,
			"wsgi.version": (1, 0),
			"wsgi.url_scheme": request.protocol,
			"wsgi.input": request_body,
			"wsgi.errors": sys.stderr,
			"wsgi.multithread": False,
			"wsgi.multiprocess": True,
			"wsgi.run_once": False,
		}
		if "Content-Type" in request.headers:
			environ["CONTENT_TYPE"] = request.headers.pop("Content-Type")
		if "Content-Length" in request.headers:
			environ["CONTENT_LENGTH"] = request.headers.pop("Content-Length")
		for key, value in request.headers.items():
			environ["HTTP_" + key.replace("-", "_").upper()] = value
		return environ

	def _log(self, status_code, request):
		access_log = logging.getLogger("tornado.access")

		if status_code < 400:
			log_method = access_log.info
		elif status_code < 500:
			log_method = access_log.warning
		else:
			log_method = access_log.error
		request_time = 1000.0 * request.request_time()
		summary = request.method + " " + request.uri + " (" + \
		          request.remote_ip + ")"
		log_method("%d %s %.2fms", status_code, summary, request_time)


#~~ customized HTTP1Connection implementation


class CustomHTTPServer(tornado.httpserver.HTTPServer):
	"""
	Custom implementation of ``tornado.httpserver.HTTPServer`` that allows defining max body sizes depending on path and
	method.

	The implementation is mostly taken from ``tornado.httpserver.HTTPServer``, the only difference is the creation
	of a ``CustomHTTP1ConnectionParameters`` instance instead of ``tornado.http1connection.HTTP1ConnectionParameters``
	which is supplied with the two new constructor arguments ``max_body_sizes`` and ``max_default_body_size`` and the
	creation of a ``CustomHTTP1ServerConnection`` instead of a ``tornado.http1connection.HTTP1ServerConnection`` upon
	connection by a client.

	``max_body_sizes`` is expected to be an iterable containing tuples of the form (method, path regex, maximum body size),
	with method and path regex having to match in order for maximum body size to take affect.

	``default_max_body_size`` is the default maximum body size to apply if no specific one from ``max_body_sizes`` matches.
	"""
	def __init__(self, *args, **kwargs):
		pass

	def initialize(self, *args, **kwargs):
		default_max_body_size = kwargs.pop("default_max_body_size", None)
		max_body_sizes = kwargs.pop("max_body_sizes", None)

		tornado.httpserver.HTTPServer.initialize(self, *args, **kwargs)

		additional = dict(default_max_body_size=default_max_body_size,
		                  max_body_sizes=max_body_sizes)
		self.conn_params = CustomHTTP1ConnectionParameters.from_stock_params(self.conn_params, **additional)


	def handle_stream(self, stream, address):
		context = tornado.httpserver._HTTPRequestContext(stream, address,
		                                                 self.protocol,
		                                                 self.trusted_downstream)
		conn = CustomHTTP1ServerConnection(stream, self.conn_params, context)
		self._connections.add(conn)
		conn.start_serving(self)


class CustomHTTP1ServerConnection(tornado.http1connection.HTTP1ServerConnection):
	"""
	A custom implementation of ``tornado.http1connection.HTTP1ServerConnection`` which utilizes a ``CustomHTTP1Connection``
	instead of a ``tornado.http1connection.HTTP1Connection`` in ``_server_request_loop``. The implementation logic is
	otherwise the same as ``tornado.http1connection.HTTP1ServerConnection``.
	"""

	@tornado.gen.coroutine
	def _server_request_loop(self, delegate):
		try:
			while True:
				conn = CustomHTTP1Connection(self.stream, False,
				                             self.params, self.context)
				request_delegate = delegate.start_request(self, conn)
				try:
					ret = yield conn.read_response(request_delegate)
				except (tornado.iostream.StreamClosedError,
						tornado.iostream.UnsatisfiableReadError):
					return
				except tornado.http1connection._QuietException:
					# This exception was already logged.
					conn.close()
					return
				except Exception:
					tornado.http1connection.gen_log.error("Uncaught exception", exc_info=True)
					conn.close()
					return
				if not ret:
					return
				yield tornado.gen.moment
		finally:
			delegate.on_close(self)


class CustomHTTP1Connection(tornado.http1connection.HTTP1Connection):
	"""
	A custom implementation of ``tornado.http1connection.HTTP1Connection`` which upon checking the ``Content-Length`` of
	the request against the configured maximum utilizes ``max_body_sizes`` and ``default_max_body_size`` as a fallback.
	"""

	def __init__(self, stream, is_client, params=None, context=None):
		if params is None:
			params = CustomHTTP1ConnectionParameters()

		tornado.http1connection.HTTP1Connection.__init__(self, stream, is_client, params=params, context=context)

		import re
		self._max_body_sizes = list(map(lambda x: (x[0], re.compile(x[1]), x[2]), self.params.max_body_sizes or list()))
		self._default_max_body_size = self.params.default_max_body_size or self.stream.max_buffer_size

	def _read_body(self, code, headers, delegate):
		"""
		Basically the same as ``tornado.http1connection.HTTP1Connection._read_body``, but determines the maximum
		content length individually for the request (utilizing ``._get_max_content_length``).

		If the individual max content length is 0 or smaller no content length is checked. If the content length of the
		current request exceeds the individual max content length, the request processing is aborted and an
		``HTTPInputError`` is raised.
		"""
		if "Content-Length" in headers:
			if "Transfer-Encoding" in headers:
				# Response cannot contain both Content-Length and
				# Transfer-Encoding headers.
				# http://tools.ietf.org/html/rfc7230#section-3.3.3
				raise tornado.httputil.HTTPInputError(
					"Response with both Transfer-Encoding and Content-Length")
			if "," in headers["Content-Length"]:
				# Proxies sometimes cause Content-Length headers to get
				# duplicated.  If all the values are identical then we can
				# use them but if they differ it's an error.
				pieces = re.split(r',\s*', headers["Content-Length"])
				if any(i != pieces[0] for i in pieces):
					raise tornado.httputil.HTTPInputError(
						"Multiple unequal Content-Lengths: %r" %
						headers["Content-Length"])
				headers["Content-Length"] = pieces[0]

			try:
				content_length = int(headers["Content-Length"])
			except ValueError:
				# Handles non-integer Content-Length value.
				raise tornado.httputil.HTTPInputError(
					"Only integer Content-Length is allowed: %s" % headers["Content-Length"])

			max_content_length = self._get_max_content_length(self._request_start_line.method, self._request_start_line.path)
			if max_content_length is not None and 0 <= max_content_length < content_length:
				raise tornado.httputil.HTTPInputError("Content-Length too long")
		else:
			content_length = None

		if code == 204:
			# This response code is not allowed to have a non-empty body,
			# and has an implicit length of zero instead of read-until-close.
			# http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.3
			if ("Transfer-Encoding" in headers or
						content_length not in (None, 0)):
				raise tornado.httputil.HTTPInputError(
					"Response with code %d should not have body" % code)
			content_length = 0

		if content_length is not None:
			return self._read_fixed_body(content_length, delegate)
		if headers.get("Transfer-Encoding") == "chunked":
			return self._read_chunked_body(delegate)
		if self.is_client:
			return self._read_body_until_close(delegate)
		return None

	def _get_max_content_length(self, method, path):
		"""
		Gets the max content length for the given method and path. Checks whether method and path match against any
		of the specific maximum content lengths supplied in ``max_body_sizes`` and returns that as the maximum content
		length if available, otherwise returns ``default_max_body_size``.

		:param method: method of the request to match against
		:param path: path of the request to match against
		:return: determine maximum content length to apply to this request, max return 0 for unlimited allowed content
		         length
		"""

		for m, p, s in self._max_body_sizes:
			if method == m and p.match(path):
				return s
		return self._default_max_body_size


class CustomHTTP1ConnectionParameters(tornado.http1connection.HTTP1ConnectionParameters):
	"""
	An implementation of ``tornado.http1connection.HTTP1ConnectionParameters`` that adds two new parameters
	``max_body_sizes`` and ``default_max_body_size``.

	For a description of these please see the documentation of ``CustomHTTPServer`` above.
	"""

	def __init__(self, *args, **kwargs):
		max_body_sizes = kwargs.pop("max_body_sizes", list())
		default_max_body_size = kwargs.pop("default_max_body_size", None)

		tornado.http1connection.HTTP1ConnectionParameters.__init__(self, *args, **kwargs)

		self.max_body_sizes = max_body_sizes
		self.default_max_body_size = default_max_body_size

	@classmethod
	def from_stock_params(cls, other, **additional):
		kwargs = dict(other.__dict__)
		for key, value in additional.items():
			kwargs[key] = value
		return cls(**kwargs)

#~~ customized large response handler


class LargeResponseHandler(RequestlessExceptionLoggingMixin,
                           CorsSupportMixin,
                           tornado.web.StaticFileHandler):
	"""
	Customized `tornado.web.StaticFileHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#tornado.web.StaticFileHandler>`_
	that allows delivery of the requested resource as attachment and access and request path validation through
	optional callbacks. Note that access validation takes place before path validation.

	Arguments:
	   path (str): The system path from which to serve files (this will be forwarded to the ``initialize`` method of
	       :class:``~tornado.web.StaticFileHandler``)
	   default_filename (str): The default filename to serve if none is explicitly specified and the request references
	       a subdirectory of the served path (this will be forwarded to the ``initialize`` method of
	       :class:``~tornado.web.StaticFileHandler`` as the ``default_filename`` keyword parameter). Defaults to ``None``.
	   as_attachment (bool): Whether to serve requested files with ``Content-Disposition: attachment`` header (``True``)
	       or not. Defaults to ``False``.
	   allow_client_caching (bool): Whether to allow the client to cache (by not setting any ``Cache-Control`` or
	       ``Expires`` headers on the response) or not.
	   access_validation (function): Callback to call in the ``get`` method to validate access to the resource. Will
	       be called with ``self.request`` as parameter which contains the full tornado request object. Should raise
	       a ``tornado.web.HTTPError`` if access is not allowed in which case the request will not be further processed.
	       Defaults to ``None`` and hence no access validation being performed.
	   path_validation (function): Callback to call in the ``get`` method to validate the requested path. Will be called
	       with the requested path as parameter. Should raise a ``tornado.web.HTTPError`` (e.g. an 404) if the requested
	       path does not pass validation in which case the request will not be further processed.
	       Defaults to ``None`` and hence no path validation being performed.
	   etag_generator (function): Callback to call for generating the value of the ETag response header. Will be
	       called with the response handler as parameter. May return ``None`` to prevent the ETag response header
	       from being set. If not provided the last modified time of the file in question will be used as returned
	       by ``get_content_version``.
	"""

	def initialize(self, path, default_filename=None, as_attachment=False, allow_client_caching=True,
	               access_validation=None, path_validation=None, etag_generator=None, name_generator=None,
	               mime_type_guesser=None):
		tornado.web.StaticFileHandler.initialize(self, os.path.abspath(path), default_filename)
		self._as_attachment = as_attachment
		self._allow_client_caching = allow_client_caching
		self._access_validation = access_validation
		self._path_validation = path_validation
		self._etag_generator = etag_generator
		self._name_generator = name_generator
		self._mime_type_guesser = mime_type_guesser

	def get(self, path, include_body=True):
		if self._access_validation is not None:
			self._access_validation(self.request)
		if self._path_validation is not None:
			self._path_validation(path)

		if "cookie" in self.request.arguments:
			self.set_cookie(self.request.arguments["cookie"][0], "true", path="/")

		result = tornado.web.StaticFileHandler.get(self, path, include_body=include_body)
		return result

	def set_extra_headers(self, path):
		if self._as_attachment:
			filename = None
			if callable(self._name_generator):
				filename = self._name_generator(path)
			if filename is None:
				filename = os.path.basename(path)

			filename = tornado.escape.url_escape(filename, plus=False)
			self.set_header("Content-Disposition", "attachment; filename=\"{}\"; filename*=UTF-8''{}".format(filename,
			                                                                                                 filename))

		if not self._allow_client_caching:
			self.set_header("Cache-Control", "max-age=0, must-revalidate, private")
			self.set_header("Expires", "-1")

	def compute_etag(self):
		if self._etag_generator is not None:
			return self._etag_generator(self)
		else:
			return self.get_content_version(self.absolute_path)

	def get_content_type(self):
		if self._mime_type_guesser is not None:
			type = self._mime_type_guesser(self.absolute_path)
			if type is not None:
				return type

		return tornado.web.StaticFileHandler.get_content_type(self)

	@classmethod
	def get_content_version(cls, abspath):
		import os
		import stat
		return os.stat(abspath)[stat.ST_MTIME]


##~~ URL Forward Handler for forwarding requests to a preconfigured static URL


class UrlProxyHandler(RequestlessExceptionLoggingMixin,
                      CorsSupportMixin,
                      tornado.web.RequestHandler):
	"""
	`tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ that proxies
	requests to a preconfigured url and returns the response. Allows delivery of the requested content as attachment
	and access validation through an optional callback.

	This will use `tornado.httpclient.AsyncHTTPClient <http://tornado.readthedocs.org/en/branch4.0/httpclient.html#tornado.httpclient.AsyncHTTPClient>`_
	for making the request to the configured endpoint and return the body of the client response with the status code
	from the client response and the following headers:

	  * ``Date``, ``Cache-Control``, ``Expires``, ``ETag``, ``Server``, ``Content-Type`` and ``Location`` will be copied over.
	  * If ``as_attachment`` is set to True, ``Content-Disposition`` will be set to ``attachment``. If ``basename`` is
	    set including the attachment's ``filename`` attribute will be set to the base name followed by the extension
	    guessed based on the MIME type from the ``Content-Type`` header of the response. If no extension can be guessed
	    no ``filename`` attribute will be set.

	Arguments:
	   url (str): URL to forward any requests to. A 404 response will be returned if this is not set. Defaults to ``None``.
	   as_attachment (bool): Whether to serve files with ``Content-Disposition: attachment`` header (``True``)
	       or not. Defaults to ``False``.
	   basename (str): base name of file names to return as part of the attachment header, see above. Defaults to ``None``.
	   access_validation (function): Callback to call in the ``get`` method to validate access to the resource. Will
	       be called with ``self.request`` as parameter which contains the full tornado request object. Should raise
	       a ``tornado.web.HTTPError`` if access is not allowed in which case the request will not be further processed.
	       Defaults to ``None`` and hence no access validation being performed.
	"""

	def initialize(self, url=None, as_attachment=False, basename=None, access_validation=None):
		tornado.web.RequestHandler.initialize(self)
		self._url = url
		self._as_attachment = as_attachment
		self._basename = basename
		self._access_validation = access_validation

	@tornado.web.asynchronous
	def get(self, *args, **kwargs):
		if self._access_validation is not None:
			self._access_validation(self.request)

		if self._url is None:
			raise tornado.web.HTTPError(404)

		client = tornado.httpclient.AsyncHTTPClient()
		r = tornado.httpclient.HTTPRequest(url=self._url, method=self.request.method, body=self.request.body, headers=self.request.headers, follow_redirects=False, allow_nonstandard_methods=True)

		try:
			return client.fetch(r, self.handle_response)
		except tornado.web.HTTPError as e:
			if hasattr(e, "response") and e.response:
				self.handle_response(e.response)
			else:
				raise tornado.web.HTTPError(500)

	def handle_response(self, response):
		if response.error and not isinstance(response.error, tornado.web.HTTPError):
			raise tornado.web.HTTPError(500)

		filename = None

		self.set_status(response.code)
		for name in ("Date", "Cache-Control", "Server", "Content-Type", "Location", "Expires", "ETag"):
			value = response.headers.get(name)
			if value:
				self.set_header(name, value)

				if name == "Content-Type":
					filename = self.get_filename(value)

		if self._as_attachment:
			if filename is not None:
				self.set_header("Content-Disposition", "attachment; filename=%s" % filename)
			else:
				self.set_header("Content-Disposition", "attachment")

		if response.body:
			self.write(response.body)
		self.finish()

	def get_filename(self, content_type):
		if not self._basename:
			return None

		typeValue = list(x.strip() for x in content_type.split(";"))
		if len(typeValue) == 0:
			return None

		extension = mimetypes.guess_extension(typeValue[0])
		if not extension:
			return None

		return "%s%s" % (self._basename, extension)


class StaticDataHandler(RequestlessExceptionLoggingMixin,
                        CorsSupportMixin,
                        tornado.web.RequestHandler):
	"""
	`tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ that returns
	static ``data`` of a configured ``content_type``.

	Arguments:
	   data (str): The data with which to respond
	   content_type (str): The content type with which to respond. Defaults to ``text/plain``
	"""

	def initialize(self, data="", content_type="text/plain"):
		self.data = data
		self.content_type = content_type

	def get(self, *args, **kwargs):
		self.set_status(200)
		self.set_header("Content-Type", self.content_type)
		self.write(self.data)
		self.flush()
		self.finish()


class DeprecatedEndpointHandler(CorsSupportMixin,
                                tornado.web.RequestHandler):
	"""
	`tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ that redirects
	to another ``url`` and logs a deprecation warning.

	Arguments:
	   url (str): URL to which to redirect
	"""

	def initialize(self, url):
		self._url = url
		self._logger = logging.getLogger(__name__)

	def _handle_method(self, *args, **kwargs):
		to_url = self._url.format(*args)
		self._logger.info("Redirecting deprecated endpoint {} to {}".format(self.request.path, to_url))
		self.redirect(to_url, permanent=True)

	# make all http methods trigger _handle_method
	get = _handle_method
	post = _handle_method
	put = _handle_method
	patch = _handle_method
	delete = _handle_method
	head = _handle_method
	options = _handle_method


class GlobalHeaderTransform(tornado.web.OutputTransform):

	HEADERS = dict()
	FORCED_HEADERS = dict()
	REMOVED_HEADERS = []

	@classmethod
	def for_headers(cls, name, headers=None, forced_headers=None, removed_headers=None):
		if headers is None:
			headers = dict()
		if forced_headers is None:
			forced_headers = dict()
		if removed_headers is None:
			removed_headers = []

		return type(octoprint.util.to_native_str(name), (GlobalHeaderTransform,),
				    dict(HEADERS=headers,
						 FORCED_HEADERS=forced_headers,
						 REMOVED_HEADERS=removed_headers))

	def __init__(self, request):
		tornado.web.OutputTransform.__init__(self, request)

	def transform_first_chunk(self, status_code, headers, chunk, finishing):
		for header, value in self.HEADERS.items():
			if not header in headers:
				headers[header] = value
		for header, value in self.FORCED_HEADERS.items():
			headers[header] = value
		for header in self.REMOVED_HEADERS:
			del headers[header]
		return status_code, headers, chunk


#~~ Factory method for creating Flask access validation wrappers from the Tornado request context


def access_validation_factory(app, validator, *args):
	"""
	Creates an access validation wrapper using the supplied validator.

	:param validator: the access validator to use inside the validation wrapper
	:return: an access validator taking a request as parameter and performing the request validation
	"""
	def f(request):
		"""
		Creates a custom wsgi and Flask request context in order to be able to process user information
		stored in the current session.

		:param request: The Tornado request for which to create the environment and context
		"""
		import flask

		wsgi_environ = WsgiInputContainer.environ(request)
		with app.request_context(wsgi_environ):
			app.session_interface.open_session(app, flask.request)
			app.login_manager.reload_user()
			validator(flask.request, *args)
	return f

def path_validation_factory(path_filter, status_code=404):
	"""
	Creates a request path validation wrapper returning the defined status code if the supplied path_filter returns False.

	:param path_filter: the path filter to use on the requested path, should return False for requests that should
	   be responded with the provided error code.
	:return: a request path validator taking a request path as parameter and performing the request validation
	"""
	def f(path):
		if not path_filter(path):
			raise tornado.web.HTTPError(status_code)
	return f

def validation_chain(*validators):
	def f(request):
		for validator in validators:
			validator(request)
	return f



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from ..exceptions import ConfigurationInvalid, CannotCheckOffline

def get_latest(target, check, full_data=False, online=True):
	python_checker = check.get("python_checker")
	if python_checker is None or not hasattr(python_checker, "get_latest"):
		raise ConfigurationInvalid("Update configuration for {} of type python_checker needs python_checker defined and have an attribute \"get_latest\"".format(target))

	if not online and not check.get("offline", False):
		raise CannotCheckOffline("{} isn't marked as 'offline' capable, but we are apparently offline right now".format(target))

	try:
		return check["python_checker"].get_latest(target, check, full_data=full_data, online=online)
	except Exception:
		import inspect
		args, _, _, _ = inspect.getargspec(check["python_checker"].get_latest)
		if "online" not in args:
			# old python_checker footprint, simply leave out the online parameter
			return check["python_checker"].get_latest(target, check, full_data=full_data)

		# some other error, raise again
		raise



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"


class Check(object):
	name = None

	def __init__(self):
		self._active = True
		self._triggered = False

	def received(self, line):
		"""Called when receiving a new line from the printer"""
		pass

	def m115(self, name, data):
		"""Called when receiving the response to an M115 from the printer"""
		pass

	def cap(self, cap, enabled):
		"""Called when receiving a capability report line"""
		pass

	@property
	def active(self):
		"""Whether this check is still active"""
		return self._active

	@property
	def triggered(self):
		"""Whether the check has been triggered"""
		return self._triggered

	def reset(self):
		self._active = True
		self._triggered = False


class AuthorCheck(Check):
	authors = ()

	AUTHOR = "| Author: ".lower()

	def received(self, line):
		if not line:
			return

		lower_line = line.lower()
		if self.AUTHOR in lower_line:
			self._triggered = any(map(lambda x: x in lower_line, self.authors))
			self._active = False




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import uuid
from octoprint.vendor.sockjs.tornado import SockJSRouter
from flask import Flask, g, request, session, Blueprint, Request, Response, current_app
from flask_login import LoginManager, current_user, session_protected, user_logged_out
from octoprint.vendor.flask_principal import Principal, Permission, RoleNeed, identity_loaded, identity_changed, UserNeed, Identity, AnonymousIdentity
from flask_babel import Babel, gettext, ngettext
from flask_assets import Environment, Bundle
from babel import Locale
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver
from collections import defaultdict, OrderedDict

from builtins import bytes, range
from past.builtins import basestring, unicode

import octoprint.util
import octoprint.util.net

from octoprint.server import util
from octoprint.util.json import JsonEncoding

import io
import os
import sys
import logging
import logging.config
import mimetypes
import atexit
import signal
import base64
import re

try:
	import fcntl
except ImportError:
	fcntl = None

SUCCESS = {}
NO_CONTENT = ("", 204)
NOT_MODIFIED = ("Not Modified", 304)

app = Flask("octoprint")

assets = None
babel = None
debug = False
safe_mode = False

printer = None
printerProfileManager = None
fileManager = None
slicingManager = None
analysisQueue = None
userManager = None
permissionManager = None
groupManager = None
eventManager = None
loginManager = None
pluginManager = None
pluginLifecycleManager = None
preemptiveCache = None
jsonEncoder = None
jsonDecoder = None
connectivityChecker = None

principals = Principal(app)

import octoprint.access.permissions as permissions
import octoprint.access.groups as groups

# we set admin_permission to a GroupPermission with the default admin group
admin_permission = octoprint.util.variable_deprecated("admin_permission has been deprecated, "
                                                      "please use individual Permissions instead",
                                                      since="1.4.0")(groups.GroupPermission(groups.ADMIN_GROUP))

# we set user_permission to a GroupPermission with the default user group
user_permission = octoprint.util.variable_deprecated("user_permission has been deprecated, "
                                                     "please use individual Permissions instead",
                                                     since="1.4.0")(groups.GroupPermission(groups.USER_GROUP))

# only import further octoprint stuff down here, as it might depend on things defined above to be initialized already
from octoprint import __version__, __branch__, __display_version__, __revision__
from octoprint.printer.profile import PrinterProfileManager
from octoprint.printer.standard import Printer
from octoprint.settings import settings
import octoprint.access.users as users
import octoprint.access.groups as groups
import octoprint.events as events
import octoprint.plugin
import octoprint.timelapse
import octoprint._version
import octoprint.filemanager.storage
import octoprint.filemanager.analysis
import octoprint.slicing
from octoprint.server.util import loginFromApiKeyRequestHandler, corsRequestHandler, \
	corsResponseHandler, requireLoginRequestHandler
from octoprint.server.util.flask import PreemptiveCache

VERSION = __version__
BRANCH = __branch__
DISPLAY_VERSION = __display_version__
REVISION = __revision__

LOCALES = []
LANGUAGES = set()

@identity_loaded.connect_via(app)
def on_identity_loaded(sender, identity):
	user = load_user(identity.id)
	if user is None:
		user = userManager.anonymous_user_factory()

	identity.provides.add(UserNeed(user.get_id()))
	for need in user.needs:
		identity.provides.add(need)


def _clear_identity(sender):
	# Remove session keys set by Flask-Principal
	for key in ('identity.id', 'identity.name', 'identity.auth_type'):
		session.pop(key, None)

	# switch to anonymous identity
	identity_changed.send(sender, identity=AnonymousIdentity())


@session_protected.connect_via(app)
def on_session_protected(sender):
	# session was deleted by session protection, that means the user is no more and we need to clear our identity
	if session.get("remember", None) == "clear":
		_clear_identity(sender)


@user_logged_out.connect_via(app)
def on_user_logged_out(sender, user=None):
	# user was logged out, clear identity
	_clear_identity(sender)


def load_user(id):
	if id is None or not userManager.enabled:
		return None

	if id == "_api":
		return userManager.api_user_factory()

	if session and "usersession.id" in session:
		sessionid = session["usersession.id"]
	else:
		sessionid = None

	if sessionid:
		user = userManager.find_user(userid=id, session=sessionid)
	else:
		user = userManager.find_user(userid=id)

	if user and user.is_active:
		return user

	return None

def load_user_from_request(request):
	user = None

	if settings().getBoolean(["accessControl", "trustBasicAuthentication"]):
		# Basic Authentication?
		user = util.get_user_for_authorization_header(request.headers.get('Authorization'))

	if settings().getBoolean(["accessControl", "trustRemoteUser"]):
		# Remote user header?
		user = util.get_user_for_remote_user_header(request)

	return user

def unauthorized_user():
	from flask import abort
	abort(403)


#~~ startup code


class Server(object):
	def __init__(self, settings=None, plugin_manager=None, connectivity_checker=None, environment_detector=None,
	             event_manager=None, host=None, port=None, v6_only=False, debug=False, safe_mode=False, allow_root=False,
	             octoprint_daemon=None):
		self._settings = settings
		self._plugin_manager = plugin_manager
		self._connectivity_checker = connectivity_checker
		self._environment_detector = environment_detector
		self._event_manager = event_manager
		self._host = host
		self._port = port
		self._v6_only = v6_only
		self._debug = debug
		self._safe_mode = safe_mode
		self._allow_root = allow_root
		self._octoprint_daemon = octoprint_daemon
		self._server = None

		self._logger = None

		self._lifecycle_callbacks = defaultdict(list)

		self._template_searchpaths = []

		self._intermediary_server = None

	def run(self):
		if not self._allow_root:
			self._check_for_root()

		if self._settings is None:
			self._settings = settings()

		if not self._settings.getBoolean(["server", "ignoreIncompleteStartup"]):
			self._settings.setBoolean(["server", "incompleteStartup"], True)
			self._settings.save()

		if self._plugin_manager is None:
			self._plugin_manager = octoprint.plugin.plugin_manager()

		global app
		global babel

		global printer
		global printerProfileManager
		global fileManager
		global slicingManager
		global analysisQueue
		global userManager
		global permissionManager
		global groupManager
		global eventManager
		global loginManager
		global pluginManager
		global pluginLifecycleManager
		global preemptiveCache
		global jsonEncoder
		global jsonDecoder
		global connectivityChecker
		global debug
		global safe_mode

		from tornado.ioloop import IOLoop
		from tornado.web import Application

		debug = self._debug
		safe_mode = self._safe_mode

		if self._v6_only and not octoprint.util.net.HAS_V6:
			raise RuntimeError("IPv6 only mode configured but system doesn't support IPv6")

		if self._host is None:
			host = self._settings.get(["server", "host"])
			if host is None:
				if octoprint.util.net.HAS_V6:
					host = "::"
				else:
					host = "0.0.0.0"

			self._host = host

		if ":" in self._host and not octoprint.util.net.HAS_V6:
			raise RuntimeError("IPv6 host address {!r} configured but system doesn't support IPv6".format(self._host))

		if self._port is None:
			self._port = self._settings.getInt(["server", "port"])
			if self._port is None:
				self._port = 5000

		self._logger = logging.getLogger(__name__)
		self._setup_heartbeat_logging()
		pluginManager = self._plugin_manager

		# monkey patch/fix some stuff
		util.tornado.fix_json_encode()
		util.flask.fix_flask_jsonify()

		self._setup_mimetypes()

		additional_translation_folders = []
		if not safe_mode:
			additional_translation_folders += [self._settings.getBaseFolder("translations")]
		util.flask.enable_additional_translations(additional_folders=additional_translation_folders)

		# setup app
		self._setup_app(app)

		# setup i18n
		self._setup_i18n(app)

		if self._settings.getBoolean(["serial", "log"]):
			# enable debug logging to serial.log
			logging.getLogger("SERIAL").setLevel(logging.DEBUG)

		# start the intermediary server
		self._start_intermediary_server()

		### IMPORTANT!
		###
		### Best do not start any subprocesses until the intermediary server shuts down again or they MIGHT inherit the
		### open port and prevent us from firing up Tornado later.
		###
		### The intermediary server's socket should have the CLOSE_EXEC flag (or its equivalent) set where possible, but
		### we can only do that if fcntl is available or we are on Windows, so better safe than sorry.
		###
		### See also issues #2035 and #2090

		# then initialize the plugin manager
		pluginManager.reload_plugins(startup=True, initialize_implementations=False)

		printerProfileManager = PrinterProfileManager()
		eventManager = self._event_manager

		analysis_queue_factories = dict(gcode=octoprint.filemanager.analysis.GcodeAnalysisQueue)
		analysis_queue_hooks = pluginManager.get_hooks("octoprint.filemanager.analysis.factory")
		for name, hook in analysis_queue_hooks.items():
			try:
				additional_factories = hook()
				analysis_queue_factories.update(**additional_factories)
			except Exception:
				self._logger.exception("Error while processing analysis queues from {}".format(name),
				                       extra=dict(plugin=name))
		analysisQueue = octoprint.filemanager.analysis.AnalysisQueue(analysis_queue_factories)

		slicingManager = octoprint.slicing.SlicingManager(self._settings.getBaseFolder("slicingProfiles"), printerProfileManager)

		storage_managers = dict()
		storage_managers[octoprint.filemanager.FileDestinations.LOCAL] = octoprint.filemanager.storage.LocalFileStorage(self._settings.getBaseFolder("uploads"))

		fileManager = octoprint.filemanager.FileManager(analysisQueue, slicingManager, printerProfileManager, initial_storage_managers=storage_managers)
		pluginLifecycleManager = LifecycleManager(pluginManager)
		preemptiveCache = PreemptiveCache(os.path.join(self._settings.getBaseFolder("data"), "preemptive_cache_config.yaml"))

		JsonEncoding.add_encoder(users.User, lambda obj: obj.as_dict())
		JsonEncoding.add_encoder(groups.Group, lambda obj: obj.as_dict())
		JsonEncoding.add_encoder(permissions.OctoPrintPermission, lambda obj: obj.as_dict())

		# start regular check if we are connected to the internet
		connectivityEnabled = self._settings.getBoolean(["server", "onlineCheck", "enabled"])
		connectivityInterval = self._settings.getInt(["server", "onlineCheck", "interval"])
		connectivityHost = self._settings.get(["server", "onlineCheck", "host"])
		connectivityPort = self._settings.getInt(["server", "onlineCheck", "port"])

		def on_connectivity_change(old_value, new_value):
			eventManager.fire(events.Events.CONNECTIVITY_CHANGED, payload=dict(old=old_value, new=new_value))

		connectivityChecker = self._connectivity_checker

		def on_settings_update(*args, **kwargs):
			# make sure our connectivity checker runs with the latest settings
			connectivityEnabled = self._settings.getBoolean(["server", "onlineCheck", "enabled"])
			connectivityInterval = self._settings.getInt(["server", "onlineCheck", "interval"])
			connectivityHost = self._settings.get(["server", "onlineCheck", "host"])
			connectivityPort = self._settings.getInt(["server", "onlineCheck", "port"])

			if connectivityChecker.enabled != connectivityEnabled \
					or connectivityChecker.interval != connectivityInterval \
					or connectivityChecker.host != connectivityHost \
					or connectivityChecker.port != connectivityPort:
				connectivityChecker.enabled = connectivityEnabled
				connectivityChecker.interval = connectivityInterval
				connectivityChecker.host = connectivityHost
				connectivityChecker.port = connectivityPort
				connectivityChecker.check_immediately()

		eventManager.subscribe(events.Events.SETTINGS_UPDATED, on_settings_update)

		components = dict(
			plugin_manager=pluginManager,
			printer_profile_manager=printerProfileManager,
			event_bus=eventManager,
			analysis_queue=analysisQueue,
			slicing_manager=slicingManager,
			file_manager=fileManager,
			plugin_lifecycle_manager=pluginLifecycleManager,
			preemptive_cache=preemptiveCache,
			json_encoder=jsonEncoder,
			json_decoder=jsonDecoder,
			connectivity_checker=connectivityChecker,
			environment_detector=self._environment_detector
		)

		#~~ setup access control

		# get additional permissions from plugins
		self._setup_plugin_permissions()

		# create group manager instance
		group_manager_factories = pluginManager.get_hooks("octoprint.access.groups.factory")
		for name, factory in group_manager_factories.items():
			try:
				groupManager = factory(components, self._settings)
				if groupManager is not None:
					self._logger.debug("Created group manager instance from factory {}".format(name))
					break
			except Exception:
				self._logger.exception("Error while creating group manager instance from factory {}".format(name))
		else:
			group_manager_name = self._settings.get(["accessControl", "groupManager"])
			try:
				clazz = octoprint.util.get_class(group_manager_name)
				groupManager = clazz()
			except AttributeError:
				self._logger.exception("Could not instantiate group manager {}, "
				                       "falling back to FilebasedGroupManager!".format(group_manager_name))
				groupManager = octoprint.access.groups.FilebasedGroupManager()
		components.update(dict(group_manager=groupManager))

		# create user manager instance
		user_manager_factories = pluginManager.get_hooks("octoprint.users.factory") # legacy, set first so that new wins
		user_manager_factories.update(pluginManager.get_hooks("octoprint.access.users.factory"))
		for name, factory in user_manager_factories.items():
			try:
				userManager = factory(components, self._settings)
				if userManager is not None:
					self._logger.debug("Created user manager instance from factory {}".format(name))
					break
			except Exception:
				self._logger.exception("Error while creating user manager instance from factory {}".format(name),
				                       extra=dict(plugin=name))
		else:
			user_manager_name = self._settings.get(["accessControl", "userManager"])
			try:
				clazz = octoprint.util.get_class(user_manager_name)
				userManager = clazz(groupManager)
			except Exception:
				self._logger.exception("Could not instantiate user manager {}, "
				                       "falling back to FilebasedUserManager!".format(user_manager_name))
				userManager = octoprint.access.users.FilebasedUserManager(groupManager)
			finally:
				userManager.enabled = self._settings.getBoolean(["accessControl", "enabled"])
		components.update(dict(user_manager=userManager))

		# create printer instance
		printer_factories = pluginManager.get_hooks("octoprint.printer.factory")
		for name, factory in printer_factories.items():
			try:
				printer = factory(components)
				if printer is not None:
					self._logger.debug("Created printer instance from factory {}".format(name))
					break
			except Exception:
				self._logger.exception("Error while creating printer instance from factory {}".format(name),
				                       extra=dict(plugin=name))
		else:
			printer = Printer(fileManager, analysisQueue, printerProfileManager)
		components.update(dict(printer=printer))

		def octoprint_plugin_inject_factory(name, implementation):
			"""Factory for injections for all OctoPrintPlugins"""
			if not isinstance(implementation, octoprint.plugin.OctoPrintPlugin):
				return None

			components_copy = dict(components)
			if "printer" in components:
				import wrapt
				import functools

				def tagwrap(f):
					@functools.wraps(f)
					def wrapper(*args, **kwargs):
						tags = kwargs.get("tags", set()) | {"source:plugin",
						                                    "plugin:{}".format(name)}
						kwargs["tags"] = tags
						return f(*args, **kwargs)
					setattr(wrapper, "__tagwrapped__", True)
					return wrapper

				class TaggedFuncsPrinter(wrapt.ObjectProxy):
					def __getattribute__(self, attr):
						__wrapped__ = super(TaggedFuncsPrinter, self).__getattribute__("__wrapped__")
						if attr == "__wrapped__":
							return __wrapped__

						item = getattr(__wrapped__, attr)
						if callable(item) \
								and ("tags" in item.__code__.co_varnames or "kwargs" in item.__code__.co_varnames) \
								and not getattr(item, "__tagwrapped__", False):
							return tagwrap(item)
						else:
							return item

				components_copy["printer"] = TaggedFuncsPrinter(components["printer"])

			props = dict()
			props.update(components_copy)
			props.update(dict(
				data_folder=os.path.join(self._settings.getBaseFolder("data"), name)
			))
			return props

		def settings_plugin_inject_factory(name, implementation):
			"""Factory for additional injections/initializations depending on plugin type"""
			if not isinstance(implementation, octoprint.plugin.SettingsPlugin):
				return

			default_settings_overlay = dict(plugins=dict())
			default_settings_overlay["plugins"][name] = implementation.get_settings_defaults()
			self._settings.add_overlay(default_settings_overlay, at_end=True)

			plugin_settings = octoprint.plugin.plugin_settings_for_settings_plugin(name, implementation)
			if plugin_settings is None:
				return

			return dict(settings=plugin_settings)

		def settings_plugin_config_migration_and_cleanup(identifier, implementation):
			"""Take care of migrating and cleaning up any old settings"""

			if not isinstance(implementation, octoprint.plugin.SettingsPlugin):
				return

			settings_version = implementation.get_settings_version()
			settings_migrator = implementation.on_settings_migrate

			if settings_version is not None and settings_migrator is not None:
				stored_version = implementation._settings.get_int([octoprint.plugin.SettingsPlugin.config_version_key])
				if stored_version is None or stored_version < settings_version:
					settings_migrator(settings_version, stored_version)
					implementation._settings.set_int([octoprint.plugin.SettingsPlugin.config_version_key], settings_version, force=True)

			implementation.on_settings_cleanup()
			implementation._settings.save()

			implementation.on_settings_initialized()

		custom_events_hooks = pluginManager.get_hooks("octoprint.events.register_custom_events")
		for name, hook in custom_events_hooks.items():
			try:
				result = hook()
				if isinstance(result, (list, tuple)):
					for event in result:
						constant, value = octoprint.events.Events.register_event(event, prefix="plugin_{}_".format(name))
						self._logger.debug("Registered event {} of plugin {} as Events.{} = \"{}\"".format(event, name, constant, value))
			except Exception:
				self._logger.exception("Error while retrieving custom event list from plugin {}".format(name),
				                       extra=dict(plugin=name))

		pluginManager.implementation_inject_factories=[octoprint_plugin_inject_factory,
		                                               settings_plugin_inject_factory]
		pluginManager.initialize_implementations()

		settingsPlugins = pluginManager.get_implementations(octoprint.plugin.SettingsPlugin)
		for implementation in settingsPlugins:
			try:
				settings_plugin_config_migration_and_cleanup(implementation._identifier, implementation)
			except Exception:
				self._logger.exception("Error while trying to migrate settings for "
				                       "plugin {}, ignoring it".format(implementation._identifier),
				                       extra=dict(plugin=implementation._identifier))

		pluginManager.implementation_post_inits=[settings_plugin_config_migration_and_cleanup]

		pluginManager.log_all_plugins()

		# log environment data now
		self._environment_detector.log_detected_environment()

		# initialize file manager and register it for changes in the registered plugins
		fileManager.initialize()
		pluginLifecycleManager.add_callback(["enabled", "disabled"], lambda name, plugin: fileManager.reload_plugins())

		# initialize slicing manager and register it for changes in the registered plugins
		slicingManager.initialize()
		pluginLifecycleManager.add_callback(["enabled", "disabled"], lambda name, plugin: slicingManager.reload_slicers())

		# setup jinja2
		self._setup_jinja2()

		# setup assets
		self._setup_assets()

		# configure timelapse
		octoprint.timelapse.valid_timelapse("test")
		octoprint.timelapse.configure_timelapse()

		# setup command triggers
		events.CommandTrigger(printer)
		if self._debug:
			events.DebugEventListener()

		# setup login manager
		self._setup_login_manager()

		# register API blueprint
		self._setup_blueprints()

		## Tornado initialization starts here

		ioloop = IOLoop()
		ioloop.install()

		enable_cors = settings().getBoolean(["api", "allowCrossOrigin"])

		self._router = SockJSRouter(self._create_socket_connection, "/sockjs",
		                            session_kls=util.sockjs.ThreadSafeSession,
		                            user_settings=dict(websocket_allow_origin="*" if enable_cors else "",
		                                               jsessionid=False,
		                                               sockjs_url="../../static/js/lib/sockjs.min.js"))

		upload_suffixes = dict(name=self._settings.get(["server", "uploads", "nameSuffix"]), path=self._settings.get(["server", "uploads", "pathSuffix"]))

		def mime_type_guesser(path):
			from octoprint.filemanager import get_mime_type
			return get_mime_type(path)

		def download_name_generator(path):
			metadata = fileManager.get_metadata("local", path)
			if metadata and "display" in metadata:
				return metadata["display"]

		download_handler_kwargs = dict(
			as_attachment=True,
			allow_client_caching=False
		)

		additional_mime_types=dict(mime_type_guesser=mime_type_guesser)

		##~~ Permission validators

		access_validators_from_plugins = []
		for plugin, hook in pluginManager.get_hooks("octoprint.server.http.access_validator").items():
			try:
				access_validators_from_plugins.append(util.tornado.access_validation_factory(app, hook))
			except Exception:
				self._logger.exception("Error while adding tornado access validator from plugin {}".format(plugin),
				                       extra=dict(plugin=plugin))
		access_validator = dict(access_validation=util.tornado.validation_chain(*access_validators_from_plugins))

		timelapse_validators = [util.tornado.access_validation_factory(app, util.flask.permission_validator, permissions.Permissions.TIMELAPSE_LIST),] + access_validators_from_plugins
		download_validators = [util.tornado.access_validation_factory(app, util.flask.permission_validator, permissions.Permissions.FILES_DOWNLOAD),] + access_validators_from_plugins
		log_validators = [util.tornado.access_validation_factory(app, util.flask.permission_validator, permissions.Permissions.PLUGIN_LOGGING_MANAGE),] + access_validators_from_plugins
		camera_validators = [util.tornado.access_validation_factory(app, util.flask.permission_validator, permissions.Permissions.WEBCAM),] + access_validators_from_plugins

		timelapse_permission_validator = dict(access_validation=util.tornado.validation_chain(*timelapse_validators))
		download_permission_validator = dict(access_validation=util.tornado.validation_chain(*download_validators))
		log_permission_validator = dict(access_validation=util.tornado.validation_chain(*log_validators))
		camera_permission_validator = dict(access_validation=util.tornado.validation_chain(*camera_validators))

		no_hidden_files_validator = dict(path_validation=util.tornado.path_validation_factory(lambda path: not octoprint.util.is_hidden_path(path),
		                                                                                      status_code=404))
		timelapse_path_validator = dict(path_validation=util.tornado.path_validation_factory(lambda path: not octoprint.util.is_hidden_path(path) and octoprint.timelapse.valid_timelapse(path),
		                                                                                status_code=404))

		def joined_dict(*dicts):
			if not len(dicts):
				return dict()

			joined = dict()
			for d in dicts:
				joined.update(d)
			return joined

		util.tornado.RequestlessExceptionLoggingMixin.LOG_REQUEST = debug
		util.tornado.CorsSupportMixin.ENABLE_CORS = enable_cors

		server_routes = self._router.urls + [
			# various downloads
			# .mpg and .mp4 timelapses:
			(r"/downloads/timelapse/(.*)", util.tornado.LargeResponseHandler, joined_dict(dict(path=self._settings.getBaseFolder("timelapse")),
			                                                                              timelapse_permission_validator,
			                                                                              download_handler_kwargs,
			                                                                              timelapse_path_validator)),
			(r"/downloads/files/local/(.*)", util.tornado.LargeResponseHandler, joined_dict(dict(path=self._settings.getBaseFolder("uploads"),
			                                                                                     as_attachment=True,
			                                                                                     name_generator=download_name_generator),
			                                                                                download_permission_validator,
			                                                                                download_handler_kwargs,
			                                                                                no_hidden_files_validator,
			                                                                                additional_mime_types)),
			(r"/downloads/logs/([^/]*)", util.tornado.LargeResponseHandler, joined_dict(dict(path=self._settings.getBaseFolder("logs"),
			                                                                                 mime_type_guesser=lambda *args, **kwargs: "text/plain"),
			                                                                            download_handler_kwargs,
			                                                                            log_permission_validator)),
			# camera snapshot
			(r"/downloads/camera/current", util.tornado.UrlProxyHandler, joined_dict(dict(url=self._settings.get(["webcam", "snapshot"]),
			                                                                              as_attachment=True),
			                                                                         camera_permission_validator)),
			# generated webassets
			(r"/static/webassets/(.*)", util.tornado.LargeResponseHandler, dict(path=os.path.join(self._settings.getBaseFolder("generated"), "webassets"))),

			# online indicators - text file with "online" as content and a transparent gif
			(r"/online.txt", util.tornado.StaticDataHandler, dict(data="online\n")),
			(r"/online.gif", util.tornado.StaticDataHandler, dict(data=bytes(base64.b64decode("R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7")),
			                                                      content_type="image/gif")),

			# deprecated endpoints
			(r"/api/logs", util.tornado.DeprecatedEndpointHandler, dict(url="/plugin/logging/logs")),
			(r"/api/logs/(.*)", util.tornado.DeprecatedEndpointHandler, dict(url="/plugin/logging/logs/{0}")),
		]

		# fetch additional routes from plugins
		for name, hook in pluginManager.get_hooks("octoprint.server.http.routes").items():
			try:
				result = hook(list(server_routes))
			except Exception:
				self._logger.exception("There was an error while retrieving additional "
				                       "server routes from plugin hook {name}".format(**locals()),
				                       extra=dict(plugin=name))
			else:
				if isinstance(result, (list, tuple)):
					for entry in result:
						if not isinstance(entry, tuple) or not len(entry) == 3:
							continue
						if not isinstance(entry[0], basestring):
							continue
						if not isinstance(entry[2], dict):
							continue

						route, handler, kwargs = entry
						route = r"/plugin/{name}/{route}".format(name=name, route=route if not route.startswith("/") else route[1:])

						self._logger.debug("Adding additional route {route} handled by handler {handler} and with additional arguments {kwargs!r}".format(**locals()))
						server_routes.append((route, handler, kwargs))

		headers =         {"X-Robots-Tag": "noindex, nofollow, noimageindex",
		                   "X-Content-Type-Options": "nosniff"}
		if not settings().getBoolean(["server", "allowFraming"]):
			headers["X-Frame-Options"] = "sameorigin"

		removed_headers = ["Server"]

		server_routes.append((r".*", util.tornado.UploadStorageFallbackHandler, dict(fallback=util.tornado.WsgiInputContainer(app.wsgi_app,
		                                                                                                                      headers=headers,
		                                                                                                                      removed_headers=removed_headers),
		                                                                             file_prefix="octoprint-file-upload-",
		                                                                             file_suffix=".tmp",
		                                                                             suffixes=upload_suffixes)))

		transforms = [util.tornado.GlobalHeaderTransform.for_headers("OctoPrintGlobalHeaderTransform",
		                                                             headers=headers,
		                                                             removed_headers=removed_headers)]

		self._tornado_app = Application(handlers=server_routes,
		                                transforms=transforms)
		max_body_sizes = [
			("POST", r"/api/files/([^/]*)", self._settings.getInt(["server", "uploads", "maxSize"])),
			("POST", r"/api/languages", 5 * 1024 * 1024)
		]

		# allow plugins to extend allowed maximum body sizes
		for name, hook in pluginManager.get_hooks("octoprint.server.http.bodysize").items():
			try:
				result = hook(list(max_body_sizes))
			except Exception:
				self._logger.exception("There was an error while retrieving additional "
				                       "upload sizes from plugin hook {name}".format(**locals()),
				                       extra=dict(plugin=name))
			else:
				if isinstance(result, (list, tuple)):
					for entry in result:
						if not isinstance(entry, tuple) or not len(entry) == 3:
							continue
						if not entry[0] in util.tornado.UploadStorageFallbackHandler.BODY_METHODS:
							continue
						if not isinstance(entry[2], int):
							continue

						method, route, size = entry
						route = r"/plugin/{name}/{route}".format(name=name, route=route if not route.startswith("/") else route[1:])

						self._logger.debug("Adding maximum body size of {size}B for {method} requests to {route})".format(**locals()))
						max_body_sizes.append((method, route, size))

		self._stop_intermediary_server()

		# initialize and bind the server
		trusted_downstream = self._settings.get(["server", "reverseProxy", "trustedDownstream"])
		if not isinstance(trusted_downstream, list):
			self._logger.warning("server.reverseProxy.trustedDownstream is not a list, skipping")
			trusted_downstreams = []

		server_kwargs = dict(max_body_sizes=max_body_sizes,
		                     default_max_body_size=self._settings.getInt(["server", "maxSize"]),
		                     xheaders=True,
		                     trusted_downstream=trusted_downstream)
		if sys.platform == "win32":
			# set 10min idle timeout under windows to hopefully make #2916 less likely
			server_kwargs.update(dict(idle_connection_timeout=600))

		self._server = util.tornado.CustomHTTPServer(self._tornado_app, **server_kwargs)

		listening_address = self._host
		if self._host == "::" and not self._v6_only:
			# special case - tornado only listens on v4 _and_ v6 if we use None as address
			listening_address = None

		self._server.listen(self._port, address=listening_address)

		### From now on it's ok to launch subprocesses again

		eventManager.fire(events.Events.STARTUP)

		# analysis backlog
		fileManager.process_backlog()

		# auto connect
		if self._settings.getBoolean(["serial", "autoconnect"]):
			self._logger.info("Autoconnect on startup is configured, trying to connect to the printer...")
			try:
				(port, baudrate) = self._settings.get(["serial", "port"]), self._settings.getInt(["serial", "baudrate"])
				printer_profile = printerProfileManager.get_default()
				connectionOptions = printer.__class__.get_connection_options()
				if port in connectionOptions["ports"] or port == "AUTO" or port is None:
					self._logger.info("Trying to connect to configured serial port {}".format(port))
					printer.connect(port=port, baudrate=baudrate, profile=printer_profile["id"] if "id" in printer_profile else "_default")
				else:
					self._logger.info("Could not find configured serial port {} in the system, cannot automatically connect to a non existing printer. Is it plugged in and booted up yet?")
			except Exception:
				self._logger.exception("Something went wrong while attempting to automatically connect to the printer")

		# start up watchdogs
		watched = self._settings.getBaseFolder("watched")
		watchdog_handler = util.watchdog.GcodeWatchdogHandler(fileManager, printer)
		watchdog_handler.initial_scan(watched)

		if self._settings.getBoolean(["feature", "pollWatched"]):
			# use less performant polling observer if explicitly configured
			observer = PollingObserver()
		else:
			# use os default
			observer = Observer()
		observer.schedule(watchdog_handler, watched)
		observer.start()

		# run our startup plugins
		octoprint.plugin.call_plugin(octoprint.plugin.StartupPlugin,
		                             "on_startup",
		                             args=(self._host, self._port),
		                             sorting_context="StartupPlugin.on_startup")

		def call_on_startup(name, plugin):
			implementation = plugin.get_implementation(octoprint.plugin.StartupPlugin)
			if implementation is None:
				return
			implementation.on_startup(self._host, self._port)
		pluginLifecycleManager.add_callback("enabled", call_on_startup)

		# prepare our after startup function
		def on_after_startup():
			if self._host == "::":
				if self._v6_only:
					# only v6
					self._logger.info("Listening on http://[::]:{port}".format(port=self._port))
				else:
					# all v4 and v6
					self._logger.info("Listening on http://0.0.0.0:{port} and http://[::]:{port}".format(port=self._port))
			else:
				self._logger.info("Listening on http://{}:{}".format(self._host if not ":" in self._host else "[" + self._host + "]",
				                                                     self._port))

			if safe_mode and self._settings.getBoolean(["server", "startOnceInSafeMode"]):
				self._logger.info("Server started successfully in safe mode as requested from config, removing flag")
				self._settings.setBoolean(["server", "startOnceInSafeMode"], False)
				self._settings.save()

			# now this is somewhat ugly, but the issue is the following: startup plugins might want to do things for
			# which they need the server to be already alive (e.g. for being able to resolve urls, such as favicons
			# or service xmls or the like). While they are working though the ioloop would block. Therefore we'll
			# create a single use thread in which to perform our after-startup-tasks, start that and hand back
			# control to the ioloop
			def work():
				octoprint.plugin.call_plugin(octoprint.plugin.StartupPlugin,
				                             "on_after_startup",
				                             sorting_context="StartupPlugin.on_after_startup")

				def call_on_after_startup(name, plugin):
					implementation = plugin.get_implementation(octoprint.plugin.StartupPlugin)
					if implementation is None:
						return
					implementation.on_after_startup()
				pluginLifecycleManager.add_callback("enabled", call_on_after_startup)

				# if there was a rogue plugin we wouldn't even have made it here, so remove startup triggered safe mode
				# flag again...
				self._settings.setBoolean(["server", "incompleteStartup"], False)
				self._settings.save()

				# make a backup of the current config
				self._settings.backup(ext="backup")

				# when we are through with that we also run our preemptive cache
				if settings().getBoolean(["devel", "cache", "preemptive"]):
					self._execute_preemptive_flask_caching(preemptiveCache)

			import threading
			threading.Thread(target=work).start()
		ioloop.add_callback(on_after_startup)

		# prepare our shutdown function
		def on_shutdown():
			# will be called on clean system exit and shutdown the watchdog observer and call the on_shutdown methods
			# on all registered ShutdownPlugins
			self._logger.info("Shutting down...")
			observer.stop()
			observer.join()
			eventManager.fire(events.Events.SHUTDOWN)
			octoprint.plugin.call_plugin(octoprint.plugin.ShutdownPlugin,
			                             "on_shutdown",
			                             sorting_context="ShutdownPlugin.on_shutdown")

			# wait for shutdown event to be processed, but maximally for 15s
			event_timeout = 15.0
			if eventManager.join(timeout=event_timeout):
				self._logger.warning("Event loop was still busy processing after {}s, shutting down anyhow".format(event_timeout))

			if self._octoprint_daemon is not None:
				self._logger.info("Cleaning up daemon pidfile")
				self._octoprint_daemon.terminated()

			self._logger.info("Goodbye!")
		atexit.register(on_shutdown)

		def sigterm_handler(*args, **kwargs):
			# will stop tornado on SIGTERM, making the program exit cleanly
			def shutdown_tornado():
				self._logger.debug("Shutting down tornado's IOLoop...")
				ioloop.stop()
			self._logger.debug("SIGTERM received...")
			ioloop.add_callback_from_signal(shutdown_tornado)
		signal.signal(signal.SIGTERM, sigterm_handler)

		try:
			# this is the main loop - as long as tornado is running, OctoPrint is running
			ioloop.start()
			self._logger.debug("Tornado's IOLoop stopped")
		except (KeyboardInterrupt, SystemExit):
			pass
		except Exception:
			self._logger.fatal("Now that is embarrassing... Something really really went wrong here. Please report this including the stacktrace below in OctoPrint's bugtracker. Thanks!")
			self._logger.exception("Stacktrace follows:")

	def _create_socket_connection(self, session):
		global printer, fileManager, analysisQueue, userManager, eventManager
		return util.sockjs.PrinterStateConnection(printer, fileManager, analysisQueue, userManager, groupManager,
		                                          eventManager, pluginManager, session)

	def _check_for_root(self):
		if "geteuid" in dir(os) and os.geteuid() == 0:
			exit("You should not run OctoPrint as root!")

	def _get_locale(self):
		global LANGUAGES

		if "l10n" in request.values:
			return Locale.negotiate([request.values["l10n"]], LANGUAGES)

		if "X-Locale" in request.headers:
			return Locale.negotiate([request.headers["X-Locale"]], LANGUAGES)

		if hasattr(g, "identity") and g.identity and userManager.enabled:
			userid = g.identity.id
			try:
				user_language = userManager.get_user_setting(userid, ("interface", "language"))
				if user_language is not None and not user_language == "_default":
					return Locale.negotiate([user_language], LANGUAGES)
			except octoprint.access.users.UnknownUser:
				pass

		default_language = self._settings.get(["appearance", "defaultLanguage"])
		if default_language is not None and not default_language == "_default" and default_language in LANGUAGES:
			return Locale.negotiate([default_language], LANGUAGES)

		return Locale.parse(request.accept_languages.best_match(LANGUAGES))

	def _setup_heartbeat_logging(self):
		logger = logging.getLogger(__name__ + ".heartbeat")

		def log_heartbeat():
			logger.info("Server heartbeat <3")

		interval = settings().getFloat(["server", "heartbeat"])
		logger.info("Starting server heartbeat, {}s interval".format(interval))

		timer = octoprint.util.RepeatedTimer(interval, log_heartbeat)
		timer.start()

	def _setup_app(self, app):
		from octoprint.server.util.flask import ReverseProxiedEnvironment, OctoPrintFlaskRequest, \
			OctoPrintFlaskResponse, OctoPrintJsonEncoder, OctoPrintSessionInterface

		s = settings()

		# setup octoprint's flask json serialization/deserialization
		app.json_encoder = OctoPrintJsonEncoder

		app.debug = self._debug
		app.config["TEMPLATES_AUTO_RELOAD"] = True
		app.config["JSONIFY_PRETTYPRINT_REGULAR"] = False

		secret_key = s.get(["server", "secretKey"])
		if not secret_key:
			import string
			from random import choice
			chars = string.ascii_lowercase + string.ascii_uppercase + string.digits
			secret_key = "".join(choice(chars) for _ in range(32))
			s.set(["server", "secretKey"], secret_key)
			s.save()

		app.secret_key = secret_key

		reverse_proxied = ReverseProxiedEnvironment(
			header_prefix=s.get(["server", "reverseProxy", "prefixHeader"]),
			header_scheme=s.get(["server", "reverseProxy", "schemeHeader"]),
			header_host=s.get(["server", "reverseProxy", "hostHeader"]),
			header_server=s.get(["server", "reverseProxy", "serverHeader"]),
			header_port=s.get(["server", "reverseProxy", "portHeader"]),
			prefix=s.get(["server", "reverseProxy", "prefixFallback"]),
			scheme=s.get(["server", "reverseProxy", "schemeFallback"]),
			host=s.get(["server", "reverseProxy", "hostFallback"]),
			server=s.get(["server", "reverseProxy", "serverFallback"]),
			port=s.get(["server", "reverseProxy", "portFallback"])
		)

		OctoPrintFlaskRequest.environment_wrapper = reverse_proxied
		app.request_class = OctoPrintFlaskRequest
		app.response_class = OctoPrintFlaskResponse
		app.session_interface = OctoPrintSessionInterface()

		@app.before_request
		def before_request():
			g.locale = self._get_locale()

		@app.after_request
		def after_request(response):
			# send no-cache headers with all POST responses
			if request.method == "POST":
				response.cache_control.no_cache = True
			response.headers.add("X-Clacks-Overhead", "GNU Terry Pratchett")
			return response

		from octoprint.util.jinja import MarkdownFilter
		MarkdownFilter(app)

	def _setup_i18n(self, app):
		global babel
		global LOCALES
		global LANGUAGES

		babel = Babel(app)

		def get_available_locale_identifiers(locales):
			result = set()

			# add available translations
			for locale in locales:
				result.add(locale.language)
				if locale.territory:
					# if a territory is specified, add that too
					result.add("%s_%s" % (locale.language, locale.territory))

			return result

		LOCALES = babel.list_translations()
		LANGUAGES = get_available_locale_identifiers(LOCALES)

		@babel.localeselector
		def get_locale():
			return self._get_locale()

	def _setup_jinja2(self):
		import re

		app.jinja_env.add_extension("jinja2.ext.do")
		app.jinja_env.add_extension("octoprint.util.jinja.trycatch")

		def regex_replace(s, find, replace):
			return re.sub(find, replace, s)

		html_header_regex = re.compile(r"<h(?P<number>[1-6])>(?P<content>.*?)</h(?P=number)>")
		def offset_html_headers(s, offset):
			def repl(match):
				number = int(match.group("number"))
				number += offset
				if number > 6:
					number = 6
				elif number < 1:
					number = 1
				return "<h{number}>{content}</h{number}>".format(number=number, content=match.group("content"))
			return html_header_regex.sub(repl, s)

		markdown_header_regex = re.compile(r"^(?P<hashs>#+)\s+(?P<content>.*)$", flags=re.MULTILINE)
		def offset_markdown_headers(s, offset):
			def repl(match):
				number = len(match.group("hashs"))
				number += offset
				if number > 6:
					number = 6
				elif number < 1:
					number = 1
				return "{hashs} {content}".format(hashs="#" * number, content=match.group("content"))
			return markdown_header_regex.sub(repl, s)

		html_link_regex = re.compile(r"<(?P<tag>a.*?)>(?P<content>.*?)</a>")
		def externalize_links(text):
			def repl(match):
				tag = match.group("tag")
				if not "href" in tag:
					return match.group(0)

				if not "target=" in tag and not "rel=" in tag:
					tag += " target=\"_blank\" rel=\"noreferrer noopener\""

				content = match.group("content")
				return "<{tag}>{content}</a>".format(tag=tag, content=content)
			return html_link_regex.sub(repl, text)

		single_quote_regex = re.compile("(?<!\\\\)'")
		def escape_single_quote(text):
			return single_quote_regex.sub("\\'", text)

		double_quote_regex = re.compile('(?<!\\\\)"')
		def escape_double_quote(text):
			return double_quote_regex.sub('\\"', text)

		app.jinja_env.filters["regex_replace"] = regex_replace
		app.jinja_env.filters["offset_html_headers"] = offset_html_headers
		app.jinja_env.filters["offset_markdown_headers"] = offset_markdown_headers
		app.jinja_env.filters["externalize_links"] = externalize_links
		app.jinja_env.filters["escape_single_quote"] = app.jinja_env.filters["esq"] = escape_single_quote
		app.jinja_env.filters["escape_double_quote"] = app.jinja_env.filters["edq"] = escape_double_quote

		# configure additional template folders for jinja2
		import jinja2
		import octoprint.util.jinja
		filesystem_loader = octoprint.util.jinja.FilteredFileSystemLoader([],
		                                                                  path_filter=lambda x: not octoprint.util.is_hidden_path(x))
		filesystem_loader.searchpath = self._template_searchpaths

		loaders = [app.jinja_loader, filesystem_loader]
		if octoprint.util.is_running_from_source():
			from markdown import markdown

			root = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../.."))
			allowed = ["AUTHORS.md", "SUPPORTERS.md", "THIRDPARTYLICENSES.md"]
			files = {"_data/" + name + ".html": os.path.join(root, name) for name in allowed}
			loaders.append(octoprint.util.jinja.SelectedFilesWithConversionLoader(files, conversion=markdown))

		jinja_loader = jinja2.ChoiceLoader(loaders)
		app.jinja_loader = jinja_loader

		self._register_template_plugins()

		# make sure plugin lifecycle events relevant for jinja2 are taken care of
		def template_enabled(name, plugin):
			if plugin.implementation is None or not isinstance(plugin.implementation, octoprint.plugin.TemplatePlugin):
				return
			self._register_additional_template_plugin(plugin.implementation)
		def template_disabled(name, plugin):
			if plugin.implementation is None or not isinstance(plugin.implementation, octoprint.plugin.TemplatePlugin):
				return
			self._unregister_additional_template_plugin(plugin.implementation)
		pluginLifecycleManager.add_callback("enabled", template_enabled)
		pluginLifecycleManager.add_callback("disabled", template_disabled)

	def _execute_preemptive_flask_caching(self, preemptive_cache):
		from werkzeug.test import EnvironBuilder
		import time

		# we clean up entries from our preemptive cache settings that haven't been
		# accessed longer than server.preemptiveCache.until days
		preemptive_cache_timeout = settings().getInt(["server", "preemptiveCache", "until"])
		cutoff_timestamp = time.time() - preemptive_cache_timeout * 24 * 60 * 60

		def filter_current_entries(entry):
			"""Returns True for entries younger than the cutoff date"""
			return "_timestamp" in entry and entry["_timestamp"] > cutoff_timestamp

		def filter_http_entries(entry):
			"""Returns True for entries targeting http or https."""
			return "base_url" in entry \
			       and entry["base_url"] \
			       and (entry["base_url"].startswith("http://")
			            or entry["base_url"].startswith("https://"))

		def filter_entries(entry):
			"""Combined filter."""
			filters = (filter_current_entries,
			           filter_http_entries)
			return all([f(entry) for f in filters])

		# filter out all old and non-http entries
		cache_data = preemptive_cache.clean_all_data(lambda root, entries: list(filter(filter_entries, entries)))
		if not cache_data:
			return

		def execute_caching():
			logger = logging.getLogger(__name__ + ".preemptive_cache")

			for route in sorted(cache_data.keys(), key=lambda x: (x.count("/"), x)):
				entries = reversed(sorted(cache_data[route], key=lambda x: x.get("_count", 0)))
				for kwargs in entries:
					plugin = kwargs.get("plugin", None)
					if plugin:
						try:
							plugin_info = pluginManager.get_plugin_info(plugin, require_enabled=True)
							if plugin_info is None:
								logger.info("About to preemptively cache plugin {} but it is not installed or enabled, preemptive caching makes no sense".format(plugin))
								continue

							implementation = plugin_info.implementation
							if implementation is None or not isinstance(implementation, octoprint.plugin.UiPlugin):
								logger.info("About to preemptively cache plugin {} but it is not a UiPlugin, preemptive caching makes no sense".format(plugin))
								continue
							if not implementation.get_ui_preemptive_caching_enabled():
								logger.info("About to preemptively cache plugin {} but it has disabled preemptive caching".format(plugin))
								continue
						except Exception:
							logger.exception("Error while trying to check if plugin {} has preemptive caching enabled, skipping entry")
							continue

					additional_request_data = kwargs.get("_additional_request_data", dict())
					kwargs = dict((k, v) for k, v in kwargs.items() if not k.startswith("_") and not k == "plugin")
					kwargs.update(additional_request_data)

					try:
						start = octoprint.util.monotonic_time()
						if plugin:
							logger.info("Preemptively caching {} (ui {}) for {!r}".format(route, plugin, kwargs))
						else:
							logger.info("Preemptively caching {} (ui _default) for {!r}".format(route, kwargs))

						headers = kwargs.get("headers", dict())
						headers["X-Force-View"] = plugin if plugin else "_default"
						headers["X-Preemptive-Record"] = "no"
						kwargs["headers"] = headers

						builder = EnvironBuilder(**kwargs)
						app(builder.get_environ(), lambda *a, **kw: None)

						logger.info("... done in {:.2f}s".format(octoprint.util.monotonic_time() - start))
					except Exception:
						logger.exception("Error while trying to preemptively cache {} for {!r}".format(route, kwargs))

		# asynchronous caching
		import threading
		cache_thread = threading.Thread(target=execute_caching, name="Preemptive Cache Worker")
		cache_thread.daemon = True
		cache_thread.start()

	def _register_template_plugins(self):
		template_plugins = pluginManager.get_implementations(octoprint.plugin.TemplatePlugin)
		for plugin in template_plugins:
			try:
				self._register_additional_template_plugin(plugin)
			except Exception:
				self._logger.exception("Error while trying to register templates of plugin {}, ignoring it".format(plugin._identifier))

	def _register_additional_template_plugin(self, plugin):
		folder = plugin.get_template_folder()
		if folder is not None and not folder in self._template_searchpaths:
			self._template_searchpaths.append(folder)

	def _unregister_additional_template_plugin(self, plugin):
		folder = plugin.get_template_folder()
		if folder is not None and folder in self._template_searchpaths:
			self._template_searchpaths.remove(folder)

	def _setup_blueprints(self):
		from octoprint.server.api import api
		import octoprint.server.views # do not remove or the index view won't be found

		blueprints = OrderedDict()
		blueprints["/api"] = api

		# also register any blueprints defined in BlueprintPlugins
		blueprints.update(self._prepare_blueprint_plugins())

		# and register a blueprint for serving the static files of asset plugins which are not blueprint plugins themselves
		blueprints.update(self._prepare_asset_plugins())

		# make sure all before/after_request hook results are attached as well
		self._add_plugin_request_handlers_to_blueprints(*blueprints.values())

		# register everything with the system
		for url_prefix, blueprint in blueprints.items():
			app.register_blueprint(blueprint, url_prefix=url_prefix)

	def _prepare_blueprint_plugins(self):
		blueprints = OrderedDict()

		blueprint_plugins = octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.BlueprintPlugin)
		for plugin in blueprint_plugins:
			try:
				blueprint, prefix = self._prepare_blueprint_plugin(plugin)
				blueprints[prefix] = blueprint
			except Exception:
				self._logger.exception("Error while registering blueprint of "
				                       "plugin {}, ignoring it".format(plugin._identifier),
				                       extra=dict(plugin=plugin._identifier))
				continue

		return blueprints

	def _prepare_asset_plugins(self):
		blueprints = OrderedDict()

		asset_plugins = octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.AssetPlugin)
		for plugin in asset_plugins:
			if isinstance(plugin, octoprint.plugin.BlueprintPlugin):
				continue
			try:
				blueprint, prefix = self._prepare_asset_plugin(plugin)
				blueprints[prefix] = blueprint
			except Exception:
				self._logger.exception("Error while registering assets of plugin "
				                       "{}, ignoring it".format(plugin._identifier),
				                       extra=dict(plugin=plugin._identifier))
				continue

		return blueprints

	def _prepare_blueprint_plugin(self, plugin):
		name = plugin._identifier
		blueprint = plugin.get_blueprint()
		if blueprint is None:
			return

		blueprint.before_request(corsRequestHandler)
		blueprint.before_request(loginFromApiKeyRequestHandler)
		blueprint.after_request(corsResponseHandler)

		if plugin.is_blueprint_protected():
			blueprint.before_request(requireLoginRequestHandler)

		url_prefix = "/plugin/{name}".format(name=name)
		app.register_blueprint(blueprint, url_prefix=url_prefix)

		if self._logger:
			self._logger.debug("Registered API of plugin {name} under URL prefix {url_prefix}".format(name=name, url_prefix=url_prefix))

		return blueprint, url_prefix

	def _prepare_asset_plugin(self, plugin):
		name = plugin._identifier

		url_prefix = "/plugin/{name}".format(name=name)
		blueprint = Blueprint("plugin." + name, name, static_folder=plugin.get_asset_folder())
		app.register_blueprint(blueprint, url_prefix=url_prefix)

		if self._logger:
			self._logger.debug("Registered assets of plugin {name} under URL prefix {url_prefix}".format(name=name, url_prefix=url_prefix))

		return blueprint, url_prefix

	def _add_plugin_request_handlers_to_blueprints(self, *blueprints):
		before_hooks = octoprint.plugin.plugin_manager().get_hooks("octoprint.server.api.before_request")
		after_hooks = octoprint.plugin.plugin_manager().get_hooks("octoprint.server.api.after_request")

		for name, hook in before_hooks.items():
			plugin = octoprint.plugin.plugin_manager().get_plugin(name)
			for blueprint in blueprints:
				try:
					result = hook(plugin=plugin)
					if isinstance(result, (list, tuple)):
						for h in result:
							blueprint.before_request(h)
				except Exception:
					self._logger.exception("Error processing before_request hooks from plugin {}".format(plugin),
					                       extra=dict(plugin=name))

		for name, hook in after_hooks.items():
			plugin = octoprint.plugin.plugin_manager().get_plugin(name)
			for blueprint in blueprints:
				try:
					result = hook(plugin=plugin)
					if isinstance(result, (list, tuple)):
						for h in result:
							blueprint.after_request(h)
				except Exception:
					self._logger.exception("Error processing after_request hooks from plugin {}".format(plugin),
					                       extra=dict(plugin=name))

	def _setup_mimetypes(self):
		# Safety measures for Windows... apparently the mimetypes module takes its translation from the windows
		# registry, and if for some weird reason that gets borked the reported MIME types can be all over the place.
		# Since at least in Chrome that can cause hilarious issues with JS files (refusal to run them and thus a
		# borked UI) we make sure that .js always maps to the correct application/javascript, and also throw in a
		# .css -> text/css for good measure.
		#
		# See #3367
		mimetypes.add_type("application/javascript", ".js")
		mimetypes.add_type("text/css", ".css")

	def _setup_assets(self):
		global app
		global assets
		global pluginManager

		util.flask.fix_webassets_cache()
		util.flask.fix_webassets_filtertool()

		base_folder = self._settings.getBaseFolder("generated")

		# clean the folder
		if self._settings.getBoolean(["devel", "webassets", "clean_on_startup"]):
			import shutil
			import errno

			for entry in ("webassets", ".webassets-cache"):
				path = os.path.join(base_folder, entry)

				# delete path if it exists
				if os.path.isdir(path):
					try:
						self._logger.debug("Deleting {path}...".format(**locals()))
						shutil.rmtree(path)
					except Exception:
						self._logger.exception("Error while trying to delete {path}, "
						                       "leaving it alone".format(**locals()))
						continue

				# re-create path
				self._logger.debug("Creating {path}...".format(**locals()))
				error_text = "Error while trying to re-create {path}, that might cause " \
				             "errors with the webassets cache".format(**locals())
				try:
					os.makedirs(path)
				except OSError as e:
					if e.errno == errno.EACCES:
						# that might be caused by the user still having the folder open somewhere, let's try again after
						# waiting a bit
						import time
						for n in range(3):
							time.sleep(0.5)
							self._logger.debug("Creating {path}: Retry #{retry} after {time}s".format(path=path,
							                                                                          retry=n+1,
							                                                                          time=(n + 1)*0.5))
							try:
								os.makedirs(path)
								break
							except Exception:
								if self._logger.isEnabledFor(logging.DEBUG):
									self._logger.exception("Ignored error while creating "
									                       "directory {path}".format(**locals()))
								pass
						else:
							# this will only get executed if we never did
							# successfully execute makedirs above
							self._logger.exception(error_text)
							continue
					else:
						# not an access error, so something we don't understand
						# went wrong -> log an error and stop
						self._logger.exception(error_text)
						continue
				except Exception:
					# not an OSError, so something we don't understand
					# went wrong -> log an error and stop
					self._logger.exception(error_text)
					continue

				self._logger.info("Reset webasset folder {path}...".format(**locals()))

		AdjustedEnvironment = type(Environment)(Environment.__name__,
		                                        (Environment,),
		                                        dict(resolver_class=util.flask.PluginAssetResolver))
		class CustomDirectoryEnvironment(AdjustedEnvironment):
			@property
			def directory(self):
				return base_folder

		assets = CustomDirectoryEnvironment(app)
		assets.debug = not self._settings.getBoolean(["devel", "webassets", "bundle"])

		UpdaterType = type(util.flask.SettingsCheckUpdater)(util.flask.SettingsCheckUpdater.__name__,
		                                                    (util.flask.SettingsCheckUpdater,),
		                                                    dict(updater=assets.updater))
		assets.updater = UpdaterType

		enable_gcodeviewer = self._settings.getBoolean(["gcodeViewer", "enabled"])
		preferred_stylesheet = self._settings.get(["devel", "stylesheet"])

		dynamic_core_assets = util.flask.collect_core_assets(enable_gcodeviewer=enable_gcodeviewer)
		dynamic_plugin_assets = util.flask.collect_plugin_assets(
			enable_gcodeviewer=enable_gcodeviewer,
			preferred_stylesheet=preferred_stylesheet
		)

		js_libs = [
			"js/lib/jquery/jquery.js",
			"js/lib/modernizr.custom.js",
			"js/lib/lodash.min.js",
			"js/lib/sprintf.min.js",
			"js/lib/knockout.js",
			"js/lib/knockout.mapping-latest.js",
			"js/lib/babel.js",
			"js/lib/avltree.js",
			"js/lib/bootstrap/bootstrap.js",
			"js/lib/bootstrap/bootstrap-modalmanager.js",
			"js/lib/bootstrap/bootstrap-modal.js",
			"js/lib/bootstrap/bootstrap-slider.js",
			"js/lib/bootstrap/bootstrap-tabdrop.js",
			"js/lib/jquery/jquery.ui.core.js",
			"js/lib/jquery/jquery.ui.widget.js",
			"js/lib/jquery/jquery.ui.mouse.js",
			"js/lib/jquery/jquery.flot.js",
			"js/lib/jquery/jquery.flot.time.js",
			"js/lib/jquery/jquery.flot.crosshair.js",
			"js/lib/jquery/jquery.flot.resize.js",
			"js/lib/jquery/jquery.iframe-transport.js",
			"js/lib/jquery/jquery.fileupload.js",
			"js/lib/jquery/jquery.slimscroll.min.js",
			"js/lib/jquery/jquery.qrcode.min.js",
			"js/lib/jquery/jquery.bootstrap.wizard.js",
			"js/lib/pnotify/pnotify.core.min.js",
			"js/lib/pnotify/pnotify.buttons.min.js",
			"js/lib/pnotify/pnotify.callbacks.min.js",
			"js/lib/pnotify/pnotify.confirm.min.js",
			"js/lib/pnotify/pnotify.desktop.min.js",
			"js/lib/pnotify/pnotify.history.min.js",
			"js/lib/pnotify/pnotify.mobile.min.js",
			"js/lib/pnotify/pnotify.nonblock.min.js",
			"js/lib/pnotify/pnotify.reference.min.js",
			"js/lib/pnotify/pnotify.tooltip.min.js",
			"js/lib/pnotify/pnotify.maxheight.js",
			"js/lib/moment-with-locales.min.js",
			"js/lib/pusher.color.min.js",
			"js/lib/detectmobilebrowser.js",
			"js/lib/md5.min.js",
			"js/lib/bootstrap-slider-knockout-binding.js",
			"js/lib/loglevel.min.js",
			"js/lib/sockjs.js",
			"js/lib/ResizeSensor.js"
		]
		js_client = [
			"js/app/client/base.js",
			"js/app/client/socket.js",
			"js/app/client/browser.js",
			"js/app/client/connection.js",
			"js/app/client/control.js",
			"js/app/client/files.js",
			"js/app/client/job.js",
			"js/app/client/languages.js",
			"js/app/client/printer.js",
			"js/app/client/printerprofiles.js",
			"js/app/client/settings.js",
			"js/app/client/slicing.js",
			"js/app/client/system.js",
			"js/app/client/timelapse.js",
			"js/app/client/users.js",
			"js/app/client/util.js",
			"js/app/client/wizard.js",
			"js/app/client/access.js"
		]

		css_libs = [
			"css/bootstrap.min.css",
			"css/bootstrap-modal.css",
			"css/bootstrap-slider.css",
			"css/bootstrap-tabdrop.css",
			"vendor/font-awesome-3.2.1/css/font-awesome.min.css",
			"vendor/font-awesome-4.7.0/css/font-awesome.min.css",
			"css/jquery.fileupload-ui.css",
			"css/pnotify.core.min.css",
			"css/pnotify.buttons.min.css",
			"css/pnotify.history.min.css"
		]

		# a couple of custom filters
		from octoprint.server.util.webassets import LessImportRewrite, JsDelimiterBundler, \
			SourceMapRewrite, SourceMapRemove, JsPluginBundle
		from webassets.filter import register_filter

		register_filter(LessImportRewrite)
		register_filter(SourceMapRewrite)
		register_filter(SourceMapRemove)
		register_filter(JsDelimiterBundler)

		def all_assets_for_plugins(collection):
			"""Gets all plugin assets for a dict of plugin->assets"""
			result = []
			for assets in collection.values():
				result += assets
			return result

		# -- JS --------------------------------------------------------------------------------------------------------

		js_filters = ["sourcemap_remove", "js_delimiter_bundler"]
		js_plugin_filters = ["sourcemap_remove", "js_delimiter_bundler"]

		def js_bundles_for_plugins(collection, filters=None):
			"""Produces JsPluginBundle instances that output IIFE wrapped assets"""
			result = OrderedDict()
			for plugin, assets in collection.items():
				if len(assets):
					result[plugin] = JsPluginBundle(plugin, *assets, filters=filters)
			return result

		js_core = dynamic_core_assets["js"] + \
		    all_assets_for_plugins(dynamic_plugin_assets["bundled"]["js"]) + \
		    ["js/app/dataupdater.js",
		     "js/app/helpers.js",
		     "js/app/main.js"]
		js_plugins = js_bundles_for_plugins(dynamic_plugin_assets["external"]["js"],
		                                    filters="js_delimiter_bundler")

		clientjs_core = dynamic_core_assets["clientjs"] + \
			all_assets_for_plugins(dynamic_plugin_assets["bundled"]["clientjs"])
		clientjs_plugins = js_bundles_for_plugins(dynamic_plugin_assets["external"]["clientjs"],
		                                          filters="js_delimiter_bundler")

		js_libs_bundle = Bundle(*js_libs,
		                        output="webassets/packed_libs.js",
		                        filters=",".join(js_filters))

		js_core_bundle = Bundle(*js_core,
		                        output="webassets/packed_core.js",
		                        filters=",".join(js_filters))

		if len(js_plugins) == 0:
			js_plugins_bundle = Bundle(*[])
		else:
			js_plugins_bundle = Bundle(*js_plugins.values(),
			                           output="webassets/packed_plugins.js",
			                           filters=",".join(js_plugin_filters))

		js_app_bundle = Bundle(js_plugins_bundle, js_core_bundle,
		                       output="webassets/packed_app.js",
		                       filters=",".join(js_filters))

		js_client_core_bundle = Bundle(*clientjs_core,
		                               output="webassets/packed_client_core.js",
		                               filters=",".join(js_filters))

		if len(clientjs_plugins) == 0:
			js_client_plugins_bundle = Bundle(*[])
		else:
			js_client_plugins_bundle = Bundle(*clientjs_plugins.values(),
			                                  output="webassets/packed_client_plugins.js",
			                                  filters=",".join(js_plugin_filters))

		js_client_bundle = Bundle(js_client_core_bundle, js_client_plugins_bundle,
		                          output="webassets/packed_client.js",
		                          filters=",".join(js_filters))

		# -- CSS -------------------------------------------------------------------------------------------------------

		css_filters = ["cssrewrite"]

		css_core = list(dynamic_core_assets["css"]) \
		           + all_assets_for_plugins(dynamic_plugin_assets["bundled"]["css"])
		css_plugins = list(all_assets_for_plugins(dynamic_plugin_assets["external"]["css"]))

		css_libs_bundle = Bundle(*css_libs,
		                         output="webassets/packed_libs.css",
		                         filters=",".join(css_filters))

		if len(css_core) == 0:
			css_core_bundle = Bundle(*[])
		else:
			css_core_bundle = Bundle(*css_core,
			                         output="webassets/packed_core.css",
			                         filters=",".join(css_filters))

		if len(css_plugins) == 0:
			css_plugins_bundle = Bundle(*[])
		else:
			css_plugins_bundle = Bundle(*css_plugins,
			                            output="webassets/packed_plugins.css",
			                            filters=",".join(css_filters))

		css_app_bundle = Bundle(css_core, css_plugins,
		                        output="webassets/packed_app.css",
		                        filters=",".join(css_filters))

		# -- LESS ------------------------------------------------------------------------------------------------------

		less_filters = ["cssrewrite", "less_importrewrite"]

		less_core = list(dynamic_core_assets["less"]) \
		            + all_assets_for_plugins(dynamic_plugin_assets["bundled"]["less"])
		less_plugins = all_assets_for_plugins(dynamic_plugin_assets["external"]["less"])

		if len(less_core) == 0:
			less_core_bundle = Bundle(*[])
		else:
			less_core_bundle = Bundle(*less_core,
			                          output="webassets/packed_core.less",
			                          filters=",".join(less_filters))

		if len(less_plugins) == 0:
			less_plugins_bundle = Bundle(*[])
		else:
			less_plugins_bundle = Bundle(*less_plugins,
			                             output="webassets/packed_plugins.less",
			                             filters=",".join(less_filters))

		less_app_bundle = Bundle(less_core, less_plugins,
		                         output="webassets/packed_app.less",
		                         filters=",".join(less_filters))

		# -- asset registration ----------------------------------------------------------------------------------------

		assets.register("js_libs", js_libs_bundle)
		assets.register("js_client_core", js_client_core_bundle)
		for plugin, bundle in clientjs_plugins.items():
			# register our collected clientjs plugin bundles so that they are bound to the environment
			assets.register("js_client_plugin_{}".format(plugin), bundle)
		assets.register("js_client_plugins", js_client_plugins_bundle)
		assets.register("js_client", js_client_bundle)
		assets.register("js_core", js_core_bundle)
		for plugin, bundle in js_plugins.items():
			# register our collected plugin bundles so that they are bound to the environment
			assets.register("js_plugin_{}".format(plugin), bundle)
		assets.register("js_plugins", js_plugins_bundle)
		assets.register("js_app", js_app_bundle)
		assets.register("css_libs", css_libs_bundle)
		assets.register("css_core", css_core_bundle)
		assets.register("css_plugins", css_plugins_bundle)
		assets.register("css_app", css_app_bundle)
		assets.register("less_core", less_core_bundle)
		assets.register("less_plugins", less_plugins_bundle)
		assets.register("less_app", less_app_bundle)

	def _setup_login_manager(self):
		global loginManager

		loginManager = LoginManager()

		# "strong" is incompatible to remember me, see maxcountryman/flask-login#156. It also causes issues with
		# clients toggling between IPv4 and IPv6 client addresses due to names being resolved one way or the other as
		# at least observed on a Win10 client targeting "localhost", resolved as both "127.0.0.1" and "::1"
		loginManager.session_protection = "basic"

		loginManager.user_callback = load_user
		loginManager.unauthorized_callback = unauthorized_user
		loginManager.anonymous_user = userManager.anonymous_user_factory
		loginManager.request_callback = load_user_from_request

		loginManager.init_app(app, add_context_processor=False)

	def _start_intermediary_server(self):
		try:
			# noinspection PyCompatibility
			from http.server import HTTPServer, BaseHTTPRequestHandler
		except ImportError:
			# noinspection PyCompatibility
			from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler

		import threading
		import socket

		host = self._host
		port = self._port

		class IntermediaryServerHandler(BaseHTTPRequestHandler):
			def __init__(self, rules=None, *args, **kwargs):
				if rules is None:
					rules = []
				self.rules = rules
				BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

			def do_GET(self):
				request_path = self.path
				if "?" in request_path:
					request_path = request_path[0:request_path.find("?")]

				for rule in self.rules:
					path, data, content_type = rule
					if request_path == path:
						self.send_response(200)
						if content_type:
							self.send_header("Content-Type", content_type)
						self.end_headers()
						if isinstance(data, unicode):
							data = data.encode("utf-8")
						self.wfile.write(data)
						break
				else:
					self.send_response(404)
					self.wfile.write("Not found".encode("utf-8"))

		base_path = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "static"))
		rules = [
			("/", ["intermediary.html",], "text/html"),
			("/favicon.ico", ["img", "tentacle-20x20.png"], "image/png"),
			("/intermediary.gif", bytes(base64.b64decode("R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7")), "image/gif")
		]

		def contents(args):
			path = os.path.join(base_path, *args)
			if not os.path.isfile(path):
				return ""

			with io.open(path, 'rb') as f:
				data = f.read()
			return data

		def process(rule):
			if len(rule) == 2:
				path, data = rule
				content_type = None
			else:
				path, data, content_type = rule

			if isinstance(data, (list, tuple)):
				data = contents(data)

			return path, data, content_type

		rules = list(map(process, filter(lambda rule: len(rule) == 2 or len(rule) == 3, rules)))

		HTTPServerV4 = HTTPServer

		class HTTPServerV6(HTTPServer):
			address_family = socket.AF_INET6

		class HTTPServerV6SingleStack(HTTPServerV6):
			def __init__(self, *args, **kwargs):
				HTTPServerV6.__init__(self, *args, **kwargs)

				# explicitly set V6ONLY flag - seems to be the default, but just to make sure...
				self.socket.setsockopt(octoprint.util.net.IPPROTO_IPV6, octoprint.util.net.IPV6_V6ONLY, 1)

		class HTTPServerV6DualStack(HTTPServerV6):
			def __init__(self, *args, **kwargs):
				HTTPServerV6.__init__(self, *args, **kwargs)

				# explicitly unset V6ONLY flag
				self.socket.setsockopt(octoprint.util.net.IPPROTO_IPV6, octoprint.util.net.IPV6_V6ONLY, 0)

		if ":" in host:
			# v6
			if host == "::" and not self._v6_only:
				ServerClass = HTTPServerV6DualStack
			else:
				ServerClass = HTTPServerV6SingleStack
		else:
			# v4
			ServerClass = HTTPServerV4

		if host == "::":
			if self._v6_only:
				self._logger.debug("Starting intermediary server on http://[::]:{port}".format(port=port))
			else:
				self._logger.debug("Starting intermediary server on http://0.0.0.0:{port} and http://[::]:{port}".format(port=port))
		else:
			self._logger.debug("Starting intermediary server on http://{}:{}".format(host if not ":" in host else "[" + host + "]", port))

		self._intermediary_server = ServerClass((host, port),
		                                        lambda *args, **kwargs: IntermediaryServerHandler(rules, *args, **kwargs),
		                                        bind_and_activate=False)

		# if possible, make sure our socket's port descriptor isn't handed over to subprocesses
		from octoprint.util.platform import set_close_exec
		try:
			set_close_exec(self._intermediary_server.fileno())
		except Exception:
			self._logger.exception("Error while attempting to set_close_exec on intermediary server socket")

		# then bind the server and have it serve our handler until stopped
		try:
			self._intermediary_server.server_bind()
			self._intermediary_server.server_activate()
		except Exception:
			self._intermediary_server.server_close()
			raise

		def serve():
			try:
				self._intermediary_server.serve_forever()
			except Exception:
				self._logger.exception("Error in intermediary server")

		thread = threading.Thread(target=serve)
		thread.daemon = True
		thread.start()

		self._logger.info("Intermediary server started")

	def _stop_intermediary_server(self):
		if self._intermediary_server is None:
			return
		self._logger.info("Shutting down intermediary server...")
		self._intermediary_server.shutdown()
		self._intermediary_server.server_close()
		self._logger.info("Intermediary server shut down")

	def _setup_plugin_permissions(self):
		global pluginManager

		from octoprint.access.permissions import PluginOctoPrintPermission

		key_whitelist = re.compile(r"[A-Za-z0-9_]*")

		def permission_key(plugin, definition):
			return "PLUGIN_{}_{}".format(plugin.upper(), definition["key"].upper())

		def permission_name(plugin, definition):
			return "{}: {}".format(plugin, definition["name"])

		def permission_role(plugin, role):
			return "plugin_{}_{}".format(plugin, role)

		def process_regular_permission(plugin_info, definition):
			permissions = []
			for key in definition.get("permissions", []):
				permission = octoprint.access.permissions.Permissions.find(key)

				if permission is None:
					# if there is still no permission found, postpone this - maybe it is a permission from
					# another plugin that hasn't been loaded yet
					return False

				permissions.append(permission)

			roles = definition.get("roles", [])
			description = definition.get("description", "")
			dangerous = definition.get("dangerous", False)
			default_groups = definition.get("default_groups", [])

			roles_and_permissions = [permission_role(plugin_info.key, role) for role in roles] + permissions

			key = permission_key(plugin_info.key, definition)
			permission = PluginOctoPrintPermission(permission_name(plugin_info.name, definition),
			                                       description,
			                                       plugin=plugin_info.key,
			                                       dangerous=dangerous,
			                                       default_groups=default_groups,
			                                       *roles_and_permissions)
			setattr(octoprint.access.permissions.Permissions,
			        key,
			        PluginOctoPrintPermission(permission_name(plugin_info.name, definition),
			                                  description,
			                                  plugin=plugin_info.key,
			                                  dangerous=dangerous,
			                                  default_groups=default_groups,
			                                  *roles_and_permissions))

			self._logger.info("Added new permission from plugin {}: {} (needs: {!r})".format(plugin_info.key,
			                                                                               key,
			                                                                               ", ".join(map(repr, permission.needs))))
			return True

		postponed = []

		hooks = pluginManager.get_hooks("octoprint.access.permissions")
		for name, factory in hooks.items():
			try:
				if isinstance(factory, (tuple, list)):
					additional_permissions = list(factory)
				elif callable(factory):
					additional_permissions = factory()
				else:
					raise ValueError("factory must be either a callable, tuple or list")

				if not isinstance(additional_permissions, (tuple, list)):
					raise ValueError("factory result must be either a tuple or a list of permission definition dicts")

				plugin_info = pluginManager.get_plugin_info(name)
				for p in additional_permissions:
					if not isinstance(p, dict):
						continue

					if not "key" in p or not "name" in p:
						continue

					if not key_whitelist.match(p["key"]):
						self._logger.warn("Got permission with invalid key from plugin {}: {}".format(name, p["key"]))
						continue

					if not process_regular_permission(plugin_info, p):
						postponed.append((plugin_info, p))
			except Exception:
				self._logger.exception("Error while creating permission instance/s from {}".format(name))

		# final resolution passes
		pass_number = 1
		still_postponed = []
		while len(postponed):
			start_length = len(postponed)
			self._logger.debug("Plugin permission resolution pass #{}, "
			                   "{} unresolved permissions...".format(pass_number, start_length))

			for plugin_info, definition in postponed:
				if not process_regular_permission(plugin_info, definition):
					still_postponed.append((plugin_info, definition))

			self._logger.debug("... pass #{} done, {} permissions left to resolve".format(pass_number,
			                                                                              len(still_postponed)))

			if len(still_postponed) == start_length:
				# no change, looks like some stuff is unresolvable - let's bail
				for plugin_info, definition in still_postponed:
					self._logger.warning("Unable to resolve permission from {}: {!r}".format(plugin_info.key, definition))
				break

			postponed = still_postponed
			still_postponed = []
			pass_number += 1


class LifecycleManager(object):
	def __init__(self, plugin_manager):
		self._plugin_manager = plugin_manager

		self._plugin_lifecycle_callbacks = defaultdict(list)
		self._logger = logging.getLogger(__name__)

		def wrap_plugin_event(lifecycle_event, new_handler):
			orig_handler = getattr(self._plugin_manager, "on_plugin_" + lifecycle_event)

			def handler(*args, **kwargs):
				if callable(orig_handler):
					orig_handler(*args, **kwargs)
				if callable(new_handler):
					new_handler(*args, **kwargs)

			return handler

		def on_plugin_event_factory(lifecycle_event):
			def on_plugin_event(name, plugin):
				self.on_plugin_event(lifecycle_event, name, plugin)
			return on_plugin_event

		for event in ("loaded", "unloaded", "enabled", "disabled"):
			wrap_plugin_event(event, on_plugin_event_factory(event))

	def on_plugin_event(self, event, name, plugin):
		for lifecycle_callback in self._plugin_lifecycle_callbacks[event]:
			lifecycle_callback(name, plugin)

	def add_callback(self, events, callback):
		if isinstance(events, basestring):
			events = [events]

		for event in events:
			self._plugin_lifecycle_callbacks[event].append(callback)

	def remove_callback(self, callback, events=None):
		if events is None:
			for event in self._plugin_lifecycle_callbacks:
				if callback in self._plugin_lifecycle_callbacks[event]:
					self._plugin_lifecycle_callbacks[event].remove(callback)
		else:
			if isinstance(events, basestring):
				events = [events]

			for event in events:
				if callback in self._plugin_lifecycle_callbacks[event]:
					self._plugin_lifecycle_callbacks[event].remove(callback)
